# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import base64
import csv
import errno
import hashlib
import itertools
import json
import os
import shutil
from fileinput import FileInput

from pex import hashing
from pex.common import CopyMode, is_pyc_dir, is_pyc_file, safe_mkdir, safe_open
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING, cast
from pex.util import CacheHelper
from pex.venv.virtualenv import Virtualenv
from pex.wheel import WHEEL, WheelMetadataLoadError

if TYPE_CHECKING:
    from typing import Callable, Iterable, Iterator, Optional, Protocol, Text, Tuple, Union

    import attr  # vendor:skip

    from pex.hashing import Hasher

    class CSVWriter(Protocol):
        def writerow(self, row):
            # type: (Iterable[Union[str, int]]) -> None
            pass

else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Digest(object):
    algorithm = attr.ib()  # type: str
    encoded_hash = attr.ib()  # type: str

    def new_hasher(self):
        # type: () -> Hasher
        return hashlib.new(self.algorithm)


@attr.s(frozen=True)
class Hash(object):
    @classmethod
    def create(cls, hasher):
        # type: (Hasher) -> Hash

        # The fingerprint encoding is defined for PEP-376 RECORD files as `urlsafe-base64-nopad`
        # which is fully spelled out in code in PEP-427:
        # + https://peps.python.org/pep-0376/#record
        # + https://peps.python.org/pep-0427/#appendix
        fingerprint = base64.urlsafe_b64encode(hasher.digest()).rstrip(b"=")
        return cls(value="{alg}={hash}".format(alg=hasher.name, hash=fingerprint.decode("ascii")))

    value = attr.ib()  # type: str

    def __str__(self):
        # type: () -> str
        return self.value


def find_and_replace_path_components(
    path,  # type: Text
    find,  # type: str
    replace,  # type: str
):
    # type: (...) -> Text
    """Replace components of `path` that are exactly `find` with `replace`.

    >>> find_and_replace_path_components("foo/bar/baz", "bar", "spam")
    foo/spam/baz
    >>>
    """
    if not find or not replace:
        raise ValueError(
            "Both find and replace must be non-empty strings. Given find={find!r} "
            "replace={replace!r}".format(find=find, replace=replace)
        )
    if not path:
        return path

    components = []
    head = path
    while head:
        new_head, tail = os.path.split(head)
        if new_head == head:
            components.append(head)
            break
        components.append(tail)
        head = new_head
    components.reverse()
    return os.path.join(*(replace if component == find else component for component in components))


@attr.s(frozen=True)
class InstalledFile(object):
    """The record of a single installed file from a PEP 376 RECORD file.

    See: https://www.python.org/dev/peps/pep-0376/#record
    """

    _PYTHON_VER_PLACEHOLDER = "pythonX.Y"

    @staticmethod
    def _python_ver(interpreter=None):
        # type: (Optional[PythonInterpreter]) -> str
        python = interpreter or PythonInterpreter.get()
        return "python{major}.{minor}".format(major=python.version[0], minor=python.version[1])

    @classmethod
    def normalized_path(
        cls,
        path,  # type: Text
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> Text
        return find_and_replace_path_components(
            path, cls._python_ver(interpreter=interpreter), cls._PYTHON_VER_PLACEHOLDER
        )

    @classmethod
    def denormalized_path(
        cls,
        path,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> Text
        return find_and_replace_path_components(
            path, cls._PYTHON_VER_PLACEHOLDER, cls._python_ver(interpreter=interpreter)
        )

    path = attr.ib()  # type: Text
    hash = attr.ib(default=None)  # type: Optional[Hash]
    size = attr.ib(default=None)  # type: Optional[int]


@attr.s(frozen=True)
class InstalledWheel(object):
    class LoadError(Exception):
        """Indicates an installed wheel was not loadable at a particular path."""

    _LAYOUT_JSON_FILENAME = ".layout.json"

    @classmethod
    def layout_file(cls, prefix_dir):
        # type: (str) -> str
        return os.path.join(prefix_dir, cls._LAYOUT_JSON_FILENAME)

    @classmethod
    def save(
        cls,
        prefix_dir,  # type: str
        stash_dir,  # type: str
        record_relpath,  # type: Text
        root_is_purelib,  # type: bool
    ):
        # type: (...) -> InstalledWheel

        # We currently need the installed wheel chroot hash for PEX-INFO / boot purposes. It is
        # expensive to calculate; so we do it here 1 time when saving the installed wheel.
        fingerprint = CacheHelper.dir_hash(prefix_dir, hasher=hashlib.sha256)

        layout = {
            "stash_dir": stash_dir,
            "record_relpath": record_relpath,
            "fingerprint": fingerprint,
            "root_is_purelib": root_is_purelib,
        }
        with open(cls.layout_file(prefix_dir), "w") as fp:
            json.dump(layout, fp, sort_keys=True)
        return cls(
            prefix_dir=prefix_dir,
            stash_dir=stash_dir,
            record_relpath=record_relpath,
            fingerprint=fingerprint,
            root_is_purelib=root_is_purelib,
        )

    @classmethod
    def load(cls, prefix_dir):
        # type: (str) -> InstalledWheel
        layout_file = cls.layout_file(prefix_dir)
        try:
            with open(layout_file) as fp:
                layout = json.load(fp)
        except (IOError, OSError) as e:
            raise cls.LoadError(
                "Failed to load an installed wheel layout from {layout_file}: {err}".format(
                    layout_file=layout_file, err=e
                )
            )
        if not isinstance(layout, dict):
            raise cls.LoadError(
                "The installed wheel layout file at {layout_file} must contain a single top-level "
                "object, found: {value}.".format(layout_file=layout_file, value=layout)
            )
        stash_dir = layout.get("stash_dir")
        record_relpath = layout.get("record_relpath")
        if not stash_dir or not record_relpath:
            raise cls.LoadError(
                "The installed wheel layout file at {layout_file} must contain an object with both "
                "`stash_dir` and `record_relpath` attributes, found: {value}".format(
                    layout_file=layout_file, value=layout
                )
            )

        fingerprint = layout.get("fingerprint")

        # N.B.: Caching root_is_purelib was not part of the original InstalledWheel layout data; so
        # we materialize the property if needed to support older installed wheel chroots.
        root_is_purelib = layout.get("root_is_purelib")
        if root_is_purelib is None:
            try:
                wheel = WHEEL.load(prefix_dir)
            except WheelMetadataLoadError as e:
                raise cls.LoadError(
                    "Failed to determine if installed wheel at {location} is platform-specific: "
                    "{err}".format(location=prefix_dir, err=e)
                )
            root_is_purelib = wheel.root_is_purelib

        return cls(
            prefix_dir=prefix_dir,
            stash_dir=cast(str, stash_dir),
            record_relpath=cast(str, record_relpath),
            fingerprint=cast("Optional[str]", fingerprint),
            root_is_purelib=root_is_purelib,
        )

    prefix_dir = attr.ib()  # type: str
    stash_dir = attr.ib()  # type: str
    record_relpath = attr.ib()  # type: Text
    fingerprint = attr.ib()  # type: Optional[str]
    root_is_purelib = attr.ib()  # type: bool

    def stashed_path(self, *components):
        # type: (*str) -> str
        return os.path.join(self.prefix_dir, self.stash_dir, *components)

    @staticmethod
    def create_installed_file(
        path,  # type: Text
        dest_dir,  # type: str
    ):
        # type: (...) -> InstalledFile
        hasher = hashlib.sha256()
        hashing.file_hash(path, digest=hasher)
        return InstalledFile(
            path=os.path.relpath(path, dest_dir),
            hash=Hash.create(hasher),
            size=os.stat(path).st_size,
        )

    def _create_record(
        self,
        dst,  # type: Text
        installed_files,  # type: Iterable[InstalledFile]
    ):
        # type: (...) -> None
        Record.write(
            dst=os.path.join(dst, self.record_relpath),
            installed_files=[
                # The RECORD entry should never include hash or size; so we replace any such entry
                # with an un-hashed and un-sized one.
                InstalledFile(self.record_relpath, hash=None, size=None)
                if installed_file.path == self.record_relpath
                else installed_file
                for installed_file in installed_files
            ],
        )

    def reinstall_flat(
        self,
        target_dir,  # type: str
        copy_mode=CopyMode.LINK,  # type: CopyMode.Value
    ):
        # type: (...) -> Iterator[Tuple[Text, Text]]
        """Re-installs the installed wheel in a flat target directory.

        N.B.: A record of reinstalled files is returned in the form of an iterator that must be
        consumed to drive the installation to completion.

        If there is an error re-installing a file due to it already existing in the target
        directory, the error is suppressed, and it's expected that the caller detects this by
        comparing the record of installed files against those installed previously.

        :return: An iterator over src -> dst pairs.
        """
        installed_files = [InstalledFile(self.record_relpath)]
        for src, dst in itertools.chain(
            self._reinstall_stash(dest_dir=target_dir, link=copy_mode is not CopyMode.COPY),
            self._reinstall_site_packages(target_dir, copy_mode=copy_mode),
        ):
            installed_files.append(self.create_installed_file(path=dst, dest_dir=target_dir))
            yield src, dst

        self._create_record(target_dir, installed_files)

    def reinstall_venv(
        self,
        venv,  # type: Virtualenv
        copy_mode=CopyMode.LINK,  # type: CopyMode.Value
        rel_extra_path=None,  # type: Optional[str]
    ):
        # type: (...) -> Iterator[Tuple[Text, Text]]
        """Re-installs the installed wheel in a venv.

        N.B.: A record of reinstalled files is returned in the form of an iterator that must be
        consumed to drive the installation to completion.

        If there is an error re-installing a file due to it already existing in the destination
        venv, the error is suppressed, and it's expected that the caller detects this by comparing
        the record of installed files against those installed previously.

        :return: An iterator over src -> dst pairs.
        """

        site_packages_dir = venv.purelib if self.root_is_purelib else venv.platlib
        site_packages_dir = (
            os.path.join(site_packages_dir, rel_extra_path) if rel_extra_path else site_packages_dir
        )

        installed_files = [InstalledFile(self.record_relpath)]
        for src, dst in itertools.chain(
            self._reinstall_stash(
                dest_dir=venv.venv_dir,
                interpreter=venv.interpreter,
                link=copy_mode is not CopyMode.COPY,
            ),
            self._reinstall_site_packages(site_packages_dir, copy_mode=copy_mode),
        ):
            installed_files.append(self.create_installed_file(path=dst, dest_dir=site_packages_dir))
            yield src, dst

        self._create_record(site_packages_dir, installed_files)

    def _reinstall_stash(
        self,
        dest_dir,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
        link=True,  # type: bool
    ):
        # type: (...) -> Iterator[Tuple[Text, Text]]

        stash_abs_path = os.path.join(self.prefix_dir, self.stash_dir)
        for root, dirs, files in os.walk(stash_abs_path, topdown=True, followlinks=True):
            dir_created = False
            for f in files:
                src = os.path.join(root, f)
                src_relpath = os.path.relpath(src, stash_abs_path)
                dst = InstalledFile.denormalized_path(
                    path=os.path.join(dest_dir, src_relpath), interpreter=interpreter
                )
                if not dir_created:
                    safe_mkdir(os.path.dirname(dst))
                    dir_created = True
                try:
                    # We only try to link regular files since linking a symlink on Linux can produce
                    # another symlink, which leaves open the possibility the src target could later
                    # go missing leaving the dst dangling.
                    if link and not os.path.islink(src):
                        try:
                            os.link(src, dst)
                            continue
                        except OSError as e:
                            if e.errno != errno.EXDEV:
                                raise e
                            link = False
                    shutil.copy(src, dst)
                except (IOError, OSError) as e:
                    if e.errno != errno.EEXIST:
                        raise e
                finally:
                    yield src, dst

    def _reinstall_site_packages(
        self,
        site_packages_dir,  # type: str
        copy_mode=CopyMode.LINK,  # type: CopyMode.Value
    ):
        # type: (...) -> Iterator[Tuple[Text, Text]]

        link = copy_mode is CopyMode.LINK
        for root, dirs, files in os.walk(self.prefix_dir, topdown=True, followlinks=True):
            if root == self.prefix_dir:
                dirs[:] = [d for d in dirs if not is_pyc_dir(d) and d != self.stash_dir]
                files[:] = [
                    f for f in files if not is_pyc_file(f) and f != self._LAYOUT_JSON_FILENAME
                ]

            traverse = set(dirs)
            for path, is_dir in itertools.chain(
                zip(dirs, itertools.repeat(True)), zip(files, itertools.repeat(False))
            ):
                src_entry = os.path.join(root, path)
                dst_entry = os.path.join(
                    site_packages_dir, os.path.relpath(src_entry, self.prefix_dir)
                )
                try:
                    if copy_mode is CopyMode.SYMLINK and not (
                        src_entry.endswith(".dist-info") and os.path.isdir(src_entry)
                    ):
                        dst_parent = os.path.dirname(dst_entry)
                        safe_mkdir(dst_parent)
                        rel_src = os.path.relpath(src_entry, dst_parent)
                        os.symlink(rel_src, dst_entry)
                        traverse.discard(path)
                    elif is_dir:
                        safe_mkdir(dst_entry)
                    else:
                        # We only try to link regular files since linking a symlink on Linux can
                        # produce another symlink, which leaves open the possibility the src_entry
                        # target could later go missing leaving the dst_entry dangling.
                        if link and not os.path.islink(src_entry):
                            try:
                                os.link(src_entry, dst_entry)
                                continue
                            except OSError as e:
                                if e.errno != errno.EXDEV:
                                    raise e
                                link = False
                        shutil.copy(src_entry, dst_entry)
                except (IOError, OSError) as e:
                    if e.errno != errno.EEXIST:
                        raise e
                finally:
                    if not is_dir:
                        yield src_entry, dst_entry

            dirs[:] = list(traverse)


class RecordError(Exception):
    pass


class RecordNotFoundError(RecordError):
    """Indicates a distribution's RECORD metadata could not be found."""


class UnrecognizedInstallationSchemeError(RecordError):
    """Indicates a distribution's RECORD was nested in an unrecognized installation scheme."""


@attr.s(frozen=True)
class DistInfoFile(object):
    path = attr.ib()  # type: Text
    content = attr.ib()  # type: bytes


@attr.s(frozen=True)
class Record(object):
    """Represents the PEP-376 RECORD of an installed wheel.

    See: https://www.python.org/dev/peps/pep-0376/#record
    """

    @classmethod
    def write(
        cls,
        dst,  # type: Text
        installed_files,  # type: Iterable[InstalledFile]
    ):
        # type: (...) -> None

        # The RECORD is a csv file with the path to each installed file in the 1st column.
        # See: https://www.python.org/dev/peps/pep-0376/#record
        with safe_open(dst, "w") as fp:
            csv_writer = cast(
                "CSVWriter",
                csv.writer(fp, delimiter=",", quotechar='"', lineterminator="\n"),
            )
            for installed_file in sorted(installed_files, key=lambda installed: installed.path):
                csv_writer.writerow(attr.astuple(installed_file, recurse=False))

    @classmethod
    def read(
        cls,
        lines,  # type: Union[FileInput[Text], Iterator[Text]]
        exclude=None,  # type: Optional[Callable[[Text], bool]]
    ):
        # type: (...) -> Iterator[InstalledFile]

        # The RECORD is a csv file with the path to each installed file in the 1st column.
        # See: https://www.python.org/dev/peps/pep-0376/#record
        for line, (path, fingerprint, file_size) in enumerate(
            csv.reader(lines, delimiter=",", quotechar='"'), start=1
        ):
            resolved_path = path
            if exclude and exclude(resolved_path):
                continue
            file_hash = Hash(fingerprint) if fingerprint else None
            size = int(file_size) if file_size else None
            yield InstalledFile(path=path, hash=file_hash, size=size)

    project_name = attr.ib()  # type: str
    version = attr.ib()  # type: str
    prefix_dir = attr.ib()  # type: str
    rel_base_dir = attr.ib()  # type: Text
    relative_path = attr.ib()  # type: Text



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import shutil

from pex.build_system import pep_517
from pex.common import safe_mkdir, safe_mkdtemp
from pex.pip.version import PipVersionValue
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.resolvers import Resolver
from pex.result import try_
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from testing import built_wheel, make_project

if TYPE_CHECKING:
    from typing import List, Optional, Text

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class FindLinksRepo(object):
    @classmethod
    def create(
        cls,
        path,  # type: str
        pip_version,  # type: PipVersionValue
    ):
        # type: (...) -> FindLinksRepo
        safe_mkdir(path, clean=True)
        return cls(path=path, resolver=ConfiguredResolver.version(pip_version))

    path = attr.ib()  # type: str
    resolver = attr.ib()  # type: Resolver

    def host(self, distribution):
        # type: (Text) -> None
        shutil.copy(distribution, os.path.join(self.path, os.path.basename(distribution)))

    def make_wheel(
        self,
        project_name,  # type: str
        version,  # type: str
        install_reqs=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> None
        with built_wheel(
            name=project_name, version=version, universal=True, install_reqs=install_reqs
        ) as wheel:
            self.host(wheel)

    def make_sdist(
        self,
        project_name,  # type: str
        version,  # type: str
        install_reqs=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> None
        with make_project(name=project_name, version=version, install_reqs=install_reqs) as project:
            self.host(
                try_(
                    pep_517.build_sdist(
                        project_directory=project,
                        dist_dir=safe_mkdtemp(),
                        target=LocalInterpreter.create(),
                        resolver=self.resolver,
                    )
                )
            )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.pip.version import PipVersion


def test_latest():
    # type: () -> None

    assert PipVersion.LATEST != PipVersion.VENDORED
    assert PipVersion.LATEST >= PipVersion.v23_1
    assert (
        max(
            (version for version in PipVersion.values() if not version.hidden),
            key=lambda pv: pv.version,
        )
        is PipVersion.LATEST
    )



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import hashlib
import os
from textwrap import dedent

from pex import pex_warnings, third_party
from pex.atomic_directory import atomic_directory
from pex.common import pluralize, safe_mkdtemp
from pex.dist_metadata import Requirement
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.pex import PEX
from pex.pex_bootstrapper import ensure_venv
from pex.pip.tool import Pip, PipVenv
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.resolvers import Resolver
from pex.result import Error, try_
from pex.targets import LocalInterpreter, RequiresPythonError, Targets
from pex.third_party import isolated
from pex.typing import TYPE_CHECKING
from pex.util import named_temporary_file
from pex.variables import ENV
from pex.venv.virtualenv import InstallationChoice, Virtualenv

if TYPE_CHECKING:
    from typing import Callable, Dict, Iterator, Optional, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def _pip_installation(
    version,  # type: PipVersionValue
    iter_distribution_locations,  # type: Callable[[], Iterator[str]]
    fingerprint,  # type: str
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Pip
    pip_root = os.path.join(ENV.PEX_ROOT, "pip", str(version))
    path = os.path.join(pip_root, "pip.pex")
    pip_interpreter = interpreter or PythonInterpreter.get()
    pip_pex_path = os.path.join(path, isolated().pex_hash, fingerprint)
    with atomic_directory(pip_pex_path) as chroot:
        if not chroot.is_finalized():
            from pex.pex_builder import PEXBuilder

            isolated_pip_builder = PEXBuilder(path=chroot.work_dir)
            isolated_pip_builder.info.venv = True
            for dist_location in iter_distribution_locations():
                isolated_pip_builder.add_dist_location(dist=dist_location)
            with named_temporary_file(prefix="", suffix=".py", mode="w") as fp:
                fp.write(
                    dedent(
                        """\
                        import os
                        import runpy

                        patches_package = os.environ.pop({patches_package_env_var_name!r}, None)
                        if patches_package:
                            # Apply runtime patches to Pip to work around issues or else bend
                            # Pip to Pex's needs.
                            __import__(patches_package)

                        runpy.run_module(mod_name="pip", run_name="__main__", alter_sys=True)
                        """
                    ).format(patches_package_env_var_name=Pip._PATCHES_PACKAGE_ENV_VAR_NAME)
                )
                fp.close()
                isolated_pip_builder.set_executable(fp.name, "__pex_patched_pip__.py")
            isolated_pip_builder.freeze()
    pip_cache = os.path.join(pip_root, "pip_cache")
    pip_pex = ensure_venv(PEX(pip_pex_path, interpreter=pip_interpreter))
    pip_venv = PipVenv(venv_dir=pip_pex.venv_dir, execute_args=tuple(pip_pex.execute_args()))
    return Pip(pip=pip_venv, version=version, pip_cache=pip_cache)


def _fingerprint(requirements):
    # type: (Tuple[Requirement, ...]) -> str
    if not requirements:
        return "no-extra-requirements"
    return hashlib.sha1("\n".join(sorted(map(str, requirements))).encode("utf-8")).hexdigest()


def _vendored_installation(
    interpreter=None,  # type: Optional[PythonInterpreter]
    resolver=None,  # type: Optional[Resolver]
    extra_requirements=(),  # type: Tuple[Requirement, ...]
):
    # type: (...) -> Pip

    def expose_vendored():
        # type: () -> Iterator[str]
        return third_party.expose(("pip", "setuptools"), interpreter=interpreter)

    if not extra_requirements:
        return _pip_installation(
            version=PipVersion.VENDORED,
            iter_distribution_locations=expose_vendored,
            interpreter=interpreter,
            fingerprint=_fingerprint(extra_requirements),
        )

    if not resolver:
        raise ValueError(
            "A resolver is required to install extra {requirements} for vendored Pip: "
            "{extra_requirements}".format(
                requirements=pluralize(extra_requirements, "requirement"),
                extra_requirements=" ".join(map(str, extra_requirements)),
            )
        )

    # This indirection works around MyPy type inference failing to see that
    # `iter_distribution_locations` is only successfully defined when resolve is not None.
    extra_requirement_resolver = resolver

    def iter_distribution_locations():
        # type: () -> Iterator[str]
        for location in expose_vendored():
            yield location

        for resolved_distribution in extra_requirement_resolver.resolve_requirements(
            requirements=tuple(map(str, extra_requirements)),
            targets=Targets.from_target(LocalInterpreter.create(interpreter)),
            pip_version=PipVersion.VENDORED,
            extra_resolver_requirements=(),
        ).distributions:
            yield resolved_distribution.distribution.location

    return _pip_installation(
        version=PipVersion.VENDORED,
        iter_distribution_locations=iter_distribution_locations,
        interpreter=interpreter,
        fingerprint=_fingerprint(extra_requirements),
    )


def _bootstrap_pip(
    version,  # type: PipVersionValue
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Callable[[], Iterator[str]]

    def bootstrap_pip():
        # type: () -> Iterator[str]

        chroot = safe_mkdtemp()
        venv = Virtualenv.create(
            venv_dir=os.path.join(chroot, "pip"),
            interpreter=interpreter,
            install_pip=InstallationChoice.YES,
        )

        for req in version.requirements:
            project_name = Requirement.parse(req).name
            target_dir = os.path.join(chroot, "reqs", project_name)
            venv.interpreter.execute(["-m", "pip", "install", "--target", target_dir, req])
            yield target_dir

    return bootstrap_pip


def _resolved_installation(
    version,  # type: PipVersionValue
    resolver=None,  # type: Optional[Resolver]
    interpreter=None,  # type: Optional[PythonInterpreter]
    extra_requirements=(),  # type: Tuple[Requirement, ...]
):
    # type: (...) -> Pip
    targets = Targets.from_target(LocalInterpreter.create(interpreter))

    bootstrap_pip_version = try_(
        compatible_version(
            targets,
            PipVersion.VENDORED,
            context="Bootstrapping Pip {version}".format(version=version),
            warn=False,
        )
    )
    if bootstrap_pip_version is not PipVersion.VENDORED and not extra_requirements:
        return _pip_installation(
            version=version,
            iter_distribution_locations=_bootstrap_pip(version, interpreter=interpreter),
            interpreter=interpreter,
            fingerprint=_fingerprint(extra_requirements),
        )

    requirements = list(version.requirements)
    requirements.extend(map(str, extra_requirements))
    if not resolver:
        raise ValueError(
            "A resolver is required to install {requirements} for Pip {version}: {reqs}".format(
                requirements=pluralize(requirements, "requirement"),
                version=version,
                reqs=" ".join(map(str, extra_requirements)),
            )
        )

    def resolve_distribution_locations():
        for resolved_distribution in resolver.resolve_requirements(
            requirements=requirements,
            targets=targets,
            pip_version=bootstrap_pip_version,
            extra_resolver_requirements=(),
        ).distributions:
            yield resolved_distribution.distribution.location

    return _pip_installation(
        version=version,
        iter_distribution_locations=resolve_distribution_locations,
        interpreter=interpreter,
        fingerprint=_fingerprint(extra_requirements),
    )


@attr.s(frozen=True)
class PipInstallation(object):
    interpreter = attr.ib()  # type: PythonInterpreter
    version = attr.ib()  # type: PipVersionValue
    extra_requirements = attr.ib()  # type: Tuple[Requirement, ...]

    def check_python_applies(self):
        # type: () -> None
        if not self.version.requires_python_applies(LocalInterpreter.create(self.interpreter)):
            raise RequiresPythonError(
                "The Pip requested was {pip_requirement} but it does not work with the interpreter "
                "selected which is {python_impl} {python_version} at {python_binary}. Pip "
                "{pip_version} requires Python {requires_python}.".format(
                    pip_requirement=self.version.requirement,
                    pip_version=self.version.value,
                    python_impl=self.interpreter.identity.interpreter,
                    python_version=self.interpreter.identity.version_str,
                    python_binary=self.interpreter.binary,
                    requires_python=self.version.requires_python,
                )
            )


def validate_targets(
    targets,  # type: Targets
    version,  # type: PipVersionValue
    context,  # type: str
):
    # type: (...) -> None
    all_targets = targets.unique_targets()
    invalid_targets = [
        target for target in all_targets if not version.requires_python_applies(target)
    ]
    if invalid_targets:
        raise RequiresPythonError(
            "The Pip requested for {context} was {pip_requirement} but it does not work with "
            "{quantifier} targets selected.\n"
            "\n"
            "Pip {pip_version} requires Python {requires_python} and the following {targets_do} "
            "not apply:\n"
            "{invalid_targets}"
            "".format(
                context=context,
                pip_requirement=version.requirement,
                quantifier="any of the"
                if len(invalid_targets) == len(all_targets)
                else "{invalid} out of the {total}".format(
                    invalid=len(invalid_targets), total=len(all_targets)
                ),
                pip_version=version.value,
                requires_python=version.requires_python,
                targets_do="target does" if len(invalid_targets) == 1 else "targets do",
                invalid_targets="\n".join(
                    "{index}. {target}".format(index=index, target=target)
                    for index, target in enumerate(invalid_targets, start=1)
                ),
            )
        )


def compatible_version(
    targets,  # type: Targets
    requested_version,  # type: PipVersionValue
    context,  # type: str
    warn=True,  # type: bool
):
    # type: (...) -> Union[PipVersionValue, Error]
    try:
        validate_targets(targets, requested_version, context)
        return requested_version
    except RequiresPythonError as e:
        remaining_versions = OrderedSet([requested_version] + list(PipVersion.values()))
        remaining_versions.discard(requested_version)
        for version in remaining_versions:
            try:
                validate_targets(targets, version, context)
                if warn:
                    pex_warnings.warn(
                        "{err}\n" "\n" "Using Pip {version} instead.".format(err=e, version=version)
                    )
                return version
            except RequiresPythonError:
                continue
    return Error(
        "No supported version of Pip is compatible with the given targets:\n{targets}".format(
            targets="\n".join(
                sorted(target.render_description() for target in targets.unique_targets())
            )
        )
    )


_PIP = {}  # type: Dict[PipInstallation, Pip]


def get_pip(
    interpreter=None,
    version=None,  # type: Optional[PipVersionValue]
    resolver=None,  # type: Optional[Resolver]
    extra_requirements=(),  # type: Tuple[Requirement, ...]
):
    # type: (...) -> Pip
    """Returns a lazily instantiated global Pip object that is safe for un-coordinated use."""
    if version:
        calculated_version = version
    elif PipVersion.DEFAULT is PipVersion.VENDORED:
        calculated_version = PipVersion.VENDORED
    else:
        # If no explicit Pip version was requested, and we're using Python 3.12+, the new semantic
        # is to allow selecting the appropriate Pip for the interpreter at hand without warning.
        # This is required since Python 3.12+ do not work with the vendored Pip.
        target = LocalInterpreter.create(interpreter)
        calculated_version = try_(
            compatible_version(
                targets=Targets.from_target(target),
                requested_version=PipVersion.DEFAULT,
                context="Selecting Pip for {target}".format(target=target.render_description()),
            )
        )

    installation = PipInstallation(
        interpreter=interpreter or PythonInterpreter.get(),
        version=calculated_version,
        extra_requirements=extra_requirements,
    )
    pip = _PIP.get(installation)
    if pip is None:
        installation.check_python_applies()
        if installation.version is PipVersion.VENDORED:
            pip = _vendored_installation(
                interpreter=interpreter,
                resolver=resolver,
                extra_requirements=installation.extra_requirements,
            )
        else:
            pip = _resolved_installation(
                version=installation.version,
                resolver=resolver,
                interpreter=interpreter,
                extra_requirements=installation.extra_requirements,
            )
        _PIP[installation] = pip
    return pip



# coding=utf-8
# Copyright 2020 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import functools
import glob
import importlib
import itertools
import os
import sys
import tarfile
import zipfile
from collections import defaultdict
from contextlib import closing
from email.message import Message
from email.parser import Parser
from io import StringIO
from textwrap import dedent

from pex import pex_warnings, specifier_sets
from pex.common import open_zip, pluralize
from pex.compatibility import to_unicode
from pex.enum import Enum
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.third_party.packaging.markers import Marker
from pex.third_party.packaging.requirements import InvalidRequirement
from pex.third_party.packaging.requirements import Requirement as PackagingRequirement
from pex.third_party.packaging.specifiers import InvalidSpecifier, SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        DefaultDict,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        List,
        Optional,
        Text,
        Tuple,
        Union,
    )

    import attr  # vendor:skip

    from pex.pep_440 import ParsedVersion
else:
    from pex.third_party import attr


class MetadataError(Exception):
    """Indicates an error reading distribution metadata."""


class UnrecognizedDistributionFormat(MetadataError):
    """Indicates a distribution file is not of any recognized format."""


class AmbiguousDistributionError(MetadataError):
    """Indicates multiple distributions were detected at a given location but one was expected."""


class MetadataNotFoundError(MetadataError):
    """Indicates an expected metadata file could not be found for a given distribution."""


class InvalidMetadataError(MetadataError):
    """Indicates a metadata value that is invalid."""


def _strip_sdist_path(sdist_path):
    # type: (Text) -> Optional[Text]
    if not sdist_path.endswith((".tar.gz", ".tgz", ".tar.bz2", ".tbz2", ".tar.xz", ".txz", ".zip")):
        return None

    sdist_basename = os.path.basename(sdist_path)
    filename, _ = os.path.splitext(sdist_basename)
    if filename.endswith(".tar"):
        filename, _ = os.path.splitext(filename)
    return filename


def parse_message(message):
    # type: (bytes) -> Message
    return cast(Message, Parser().parse(StringIO(to_unicode(message))))


@attr.s(frozen=True)
class DistMetadataFile(object):
    type = attr.ib()  # type: MetadataType.Value
    location = attr.ib()  # type: Text
    rel_path = attr.ib()  # type: Text
    project_name = attr.ib()  # type: ProjectName
    version = attr.ib()  # type: Version
    pkg_info = attr.ib(eq=False)  # type: Message

    def render_description(self):
        # type: () -> str
        return "{project_name} {version} metadata from {rel_path} at {location}".format(
            project_name=self.project_name,
            version=self.version,
            rel_path=self.rel_path,
            location=self.location,
        )


@attr.s(frozen=True)
class MetadataFiles(object):
    metadata = attr.ib()  # type: DistMetadataFile
    _additional_metadata_files = attr.ib(default=())  # type: Tuple[Text, ...]
    _read_function = attr.ib(default=None)  # type: Optional[Callable[[Text], bytes]]

    def metadata_file_rel_path(self, metadata_file_name):
        # type: (Text) -> Optional[Text]
        for rel_path in self._additional_metadata_files:
            if os.path.basename(rel_path) == metadata_file_name:
                return rel_path
        return None

    def read(self, metadata_file_name):
        # type: (Text) -> Optional[bytes]
        rel_path = self.metadata_file_rel_path(metadata_file_name)
        if rel_path is None or self._read_function is None:
            return None
        return self._read_function(rel_path)


class MetadataType(Enum["MetadataType.Value"]):
    class Value(Enum.Value):
        def load_metadata(
            self,
            location,  # type: Text
            project_name=None,  # type: Optional[ProjectName]
            rescan=False,  # type: bool
        ):
            # type: (...) -> Optional[MetadataFiles]
            return load_metadata(
                location, project_name=project_name, restrict_types_to=(self,), rescan=rescan
            )

    DIST_INFO = Value(".dist-info")
    EGG_INFO = Value(".egg-info")
    PKG_INFO = Value("PKG-INFO")


@attr.s(frozen=True)
class MetadataKey(object):
    metadata_type = attr.ib()  # type: MetadataType.Value
    location = attr.ib()  # type: Text


def _find_installed_metadata_files(
    location,  # type: Text
    metadata_type,  # type: MetadataType.Value
    metadata_dir_glob,  # type: str
    metadata_file_name,  # type: Text
):
    # type: (...) -> Iterator[MetadataFiles]
    metadata_files = glob.glob(os.path.join(location, metadata_dir_glob, metadata_file_name))
    for path in metadata_files:
        with open(path, "rb") as fp:
            metadata = parse_message(fp.read())
            project_name_and_version = ProjectNameAndVersion.from_parsed_pkg_info(
                source=path, pkg_info=metadata
            )

            def read_function(rel_path):
                # type: (Text) -> bytes
                with open(os.path.join(location, rel_path), "rb") as fp:
                    return fp.read()

            yield MetadataFiles(
                metadata=DistMetadataFile(
                    type=metadata_type,
                    location=location,
                    rel_path=os.path.relpath(path, location),
                    project_name=project_name_and_version.canonicalized_project_name,
                    version=project_name_and_version.canonicalized_version,
                    pkg_info=metadata,
                ),
                additional_metadata_files=tuple(
                    os.path.relpath(metadata_path, location)
                    for metadata_path in glob.glob(os.path.join(os.path.dirname(path), "*"))
                    if os.path.basename(metadata_path) != metadata_file_name
                ),
                read_function=read_function,
            )


def find_wheel_metadata(location):
    # type: (Text) -> Optional[MetadataFiles]
    with open_zip(location) as zf:
        for name in zf.namelist():
            if name.endswith("/"):
                continue
            dist_info_dir, metadata_file = os.path.split(name)
            if os.path.dirname(dist_info_dir):
                continue
            if "METADATA" != metadata_file:
                continue

            with zf.open(name) as fp:
                metadata = parse_message(fp.read())
                project_name_and_version = ProjectNameAndVersion.from_parsed_pkg_info(
                    source=os.path.join(location, name), pkg_info=metadata
                )
                metadata_file_name = os.path.basename(name)
                files = []  # type: List[Text]
                for rel_path in zf.namelist():
                    head, tail = os.path.split(rel_path)
                    if dist_info_dir == head and tail != metadata_file_name:
                        files.append(rel_path)

                def read_function(rel_path):
                    # type: (Text) -> bytes
                    with open_zip(location) as zf:
                        return zf.read(rel_path)

                return MetadataFiles(
                    metadata=DistMetadataFile(
                        type=MetadataType.DIST_INFO,
                        location=location,
                        rel_path=name,
                        project_name=project_name_and_version.canonicalized_project_name,
                        version=project_name_and_version.canonicalized_version,
                        pkg_info=metadata,
                    ),
                    additional_metadata_files=tuple(files),
                    read_function=read_function,
                )

    return None


def _is_dist_pkg_info_file_path(file_path):
    # type: (Text) -> bool

    # N.B.: Should be: <project name>-<version>/PKG-INFO
    project_dir, metadata_file = os.path.split(file_path)
    if os.path.dirname(project_dir):
        return False
    if not "-" in project_dir:
        return False
    return "PKG-INFO" == metadata_file


def find_zip_sdist_metadata(location):
    # type: (Text) -> Optional[DistMetadataFile]
    with open_zip(location) as zf:
        for name in zf.namelist():
            if name.endswith("/") or not _is_dist_pkg_info_file_path(name):
                continue
            with zf.open(name) as fp:
                metadata = parse_message(fp.read())
                project_name_and_version = ProjectNameAndVersion.from_parsed_pkg_info(
                    source=os.path.join(location, name), pkg_info=metadata
                )
                return DistMetadataFile(
                    type=MetadataType.PKG_INFO,
                    location=location,
                    rel_path=name,
                    project_name=project_name_and_version.canonicalized_project_name,
                    version=project_name_and_version.canonicalized_version,
                    pkg_info=metadata,
                )

    return None


def find_tar_sdist_metadata(location):
    # type: (Text) -> Optional[DistMetadataFile]
    with tarfile.open(location) as tf:
        for member in tf.getmembers():
            if not member.isreg() or not _is_dist_pkg_info_file_path(member.name):
                continue

            file_obj = tf.extractfile(member)
            if file_obj is None:
                raise IOError(
                    errno.ENOENT,
                    "Could not find {rel_path} in {location}.".format(
                        rel_path=member.name, location=location
                    ),
                )
            with closing(file_obj) as fp:
                metadata = parse_message(fp.read())
                project_name_and_version = ProjectNameAndVersion.from_parsed_pkg_info(
                    source=os.path.join(location, member.name), pkg_info=metadata
                )
                return DistMetadataFile(
                    type=MetadataType.PKG_INFO,
                    location=location,
                    rel_path=member.name,
                    project_name=project_name_and_version.canonicalized_project_name,
                    version=project_name_and_version.canonicalized_version,
                    pkg_info=metadata,
                )

    return None


_METADATA_FILES = {}  # type: Dict[MetadataKey, Tuple[MetadataFiles, ...]]


def iter_metadata_files(
    location,  # type: Text
    restrict_types_to=(),  # type: Tuple[MetadataType.Value, ...]
    rescan=False,  # type: bool
):
    # type: (...) -> Iterator[MetadataFiles]

    files = []
    for metadata_type in restrict_types_to or MetadataType.values():
        key = MetadataKey(metadata_type=metadata_type, location=location)
        if rescan:
            _METADATA_FILES.pop(key, None)
        if key not in _METADATA_FILES:
            listing = []  # type: List[MetadataFiles]
            if MetadataType.DIST_INFO is metadata_type:
                if os.path.isdir(location):
                    listing.extend(
                        _find_installed_metadata_files(
                            location, MetadataType.DIST_INFO, "*.dist-info", "METADATA"
                        )
                    )
                elif location.endswith(".whl") and zipfile.is_zipfile(location):
                    metadata_files = find_wheel_metadata(location)
                    if metadata_files:
                        listing.append(metadata_files)
            elif MetadataType.EGG_INFO is metadata_type and os.path.isdir(location):
                listing.extend(
                    _find_installed_metadata_files(
                        location, MetadataType.EGG_INFO, "*.egg-info", "PKG-INFO"
                    )
                )
            elif MetadataType.PKG_INFO is metadata_type:
                if location.endswith(".zip") and zipfile.is_zipfile(location):
                    metadata_file = find_zip_sdist_metadata(location)
                    if metadata_file:
                        listing.append(MetadataFiles(metadata=metadata_file))
                elif location.endswith(
                    (".tar.gz", ".tgz", ".tar.bz2", ".tbz2", ".tar.xz", ".txz")
                ) and tarfile.is_tarfile(location):
                    metadata_file = find_tar_sdist_metadata(location)
                    if metadata_file:
                        listing.append(MetadataFiles(metadata=metadata_file))
            _METADATA_FILES[key] = tuple(listing)
        files.append(_METADATA_FILES[key])
    return itertools.chain.from_iterable(files)


def load_metadata(
    location,  # type: Text
    project_name=None,  # type: Optional[ProjectName]
    restrict_types_to=(),  # type: Tuple[MetadataType.Value, ...]
    rescan=False,  # type: bool
):
    # type: (...) -> Optional[MetadataFiles]
    all_metadata_files = [
        metadata_files
        for metadata_files in iter_metadata_files(
            location, restrict_types_to=restrict_types_to, rescan=rescan
        )
        if project_name is None or project_name == metadata_files.metadata.project_name
    ]
    if len(all_metadata_files) == 1:
        return all_metadata_files[0]
    if len(all_metadata_files) > 1:
        raise AmbiguousDistributionError(
            "Found more than one distribution inside {location}:\n{metadata_files}".format(
                location=location,
                metadata_files="\n".join(
                    metadata_file.metadata.rel_path for metadata_file in all_metadata_files
                ),
            )
        )
    return None


@attr.s(frozen=True)
class ProjectNameAndVersion(object):
    @classmethod
    def from_parsed_pkg_info(cls, source, pkg_info):
        # type: (Text, Message) -> ProjectNameAndVersion
        project_name = pkg_info.get("Name", None)
        version = pkg_info.get("Version", None)
        if project_name is None or version is None:
            raise MetadataError(
                "The 'Name' and 'Version' fields are not both present in package metadata for "
                "{source}:\n{fields}".format(
                    source=source,
                    fields="\n".join("{}: {}".format(k, v) for k, v in pkg_info.items()),
                )
            )
        return cls(project_name=pkg_info["Name"], version=pkg_info["Version"])

    @classmethod
    def from_filename(cls, path):
        # type: (Text) -> ProjectNameAndVersion
        # Handle wheels:
        #
        # The wheel filename convention is specified here:
        #   https://www.python.org/dev/peps/pep-0427/#file-name-convention.
        if path.endswith(".whl"):
            project_name, version, _ = os.path.basename(path).split("-", 2)
            return cls(project_name=project_name, version=version)

        # Handle sdists:
        #
        # The sdist name format has no accepted specification yet, but there is a proposal here:
        #   https://www.python.org/dev/peps/pep-0625/#specification.
        #
        # We do the best we can to support the current landscape. A version number can technically
        # contain a dash though, even under the standards, in un-normalized form:
        #   https://www.python.org/dev/peps/pep-0440/#pre-release-separators.
        # For those cases this logic will produce incorrect results and it does not seem there is
        # much we can do since both project names and versions can contain both alphanumeric
        # characters and dashes.
        fname = _strip_sdist_path(path)
        if fname is not None:
            components = fname.rsplit("-", 1)
            if len(components) == 2:
                project_name, version = components
                return cls(project_name=project_name, version=version)

        raise UnrecognizedDistributionFormat(
            "The distribution at path {!r} does not have a file name matching known sdist or wheel "
            "file name formats.".format(path)
        )

    project_name = attr.ib()  # type: Text
    version = attr.ib()  # type: Text

    @property
    def canonicalized_project_name(self):
        # type: () -> ProjectName
        return ProjectName(self.project_name)

    @property
    def canonicalized_version(self):
        # type: () -> Version
        return Version(self.version)


def project_name_and_version(
    location,  # type: Union[Text, Distribution, Message, MetadataFiles]
    fallback_to_filename=True,  # type: bool
):
    # type: (...) -> Optional[ProjectNameAndVersion]
    """Extracts name and version metadata from dist.

    :param location: A distribution to extract project name and version metadata from.
    :return: The project name and version.
    :raise: MetadataError if dist has invalid metadata.
    """
    if isinstance(location, Distribution):
        return ProjectNameAndVersion(project_name=location.project_name, version=location.version)
    if isinstance(location, MetadataFiles):
        return ProjectNameAndVersion(
            project_name=location.metadata.project_name.raw, version=location.metadata.version.raw
        )

    pkg_info = None  # type: Optional[Message]
    if isinstance(location, Message):
        pkg_info = location
    else:
        metadata_files = load_metadata(location)
        if metadata_files:
            pkg_info = metadata_files.metadata.pkg_info
    if pkg_info is not None:
        if isinstance(location, str):
            source = location
        else:
            source = "<parsed message>"
        return ProjectNameAndVersion.from_parsed_pkg_info(source=source, pkg_info=pkg_info)
    if fallback_to_filename and not isinstance(location, (Distribution, Message)):
        return ProjectNameAndVersion.from_filename(location)
    return None


def requires_python(location):
    # type: (Union[Distribution, MetadataFiles, Text]) -> Optional[SpecifierSet]
    """Examines dist for `Python-Requires` metadata and returns version constraints if any.

    See: https://www.python.org/dev/peps/pep-0345/#requires-python

    :param location: A distribution to check for `Python-Requires` metadata.
    :return: The required python version specifiers.
    """
    if isinstance(location, Distribution):
        return location.metadata.requires_python

    metadata_files = None  # type: Optional[MetadataFiles]
    if isinstance(location, MetadataFiles):
        metadata_files = location
    else:
        # N.B.: This load can fail, but the source is the location of the metadata files, which
        # contains useful information in the path name for identifying the problem project and
        # version.
        metadata_files = load_metadata(location)
    if metadata_files is None:
        return None

    python_requirement = metadata_files.metadata.pkg_info.get("Requires-Python", None)
    if python_requirement is None:
        return None
    try:
        return SpecifierSet(python_requirement)
    except InvalidSpecifier as e:
        raise InvalidMetadataError(
            "Invalid Requires-Python metadata found in {source} {value!r}: {err}".format(
                source=metadata_files.metadata.render_description(), value=python_requirement, err=e
            )
        )


def _parse_requires_txt(content):
    # type: (bytes) -> Iterator[Union[Requirement, Tuple[int, Text, RequirementParseError]]]
    # See:
    # + High level: https://setuptools.pypa.io/en/latest/deprecated/python_eggs.html#requires-txt
    # + Low level:
    #   + https://github.com/pypa/setuptools/blob/fbe0d7962822c2a1fdde8dd179f2f8b8c8bf8892/pkg_resources/__init__.py#L3256-L3279
    #   + https://github.com/pypa/setuptools/blob/fbe0d7962822c2a1fdde8dd179f2f8b8c8bf8892/pkg_resources/__init__.py#L2792-L2818
    marker = ""
    for line_no, line in enumerate(content.decode("utf-8").splitlines(), start=1):
        line = line.strip()
        if not line:
            continue
        if line.startswith("[") and line.endswith("]"):
            section = line[1:-1]
            extra, _, mark = section.partition(":")
            markers = []  # type: List[Text]
            if extra:
                markers.append('extra == "{extra}"'.format(extra=extra))
            if mark:
                markers.append(mark)
            if markers:
                marker = "; {markers}".format(markers=" and ".join(markers))
        else:
            req = line + marker
            try:
                yield Requirement.parse(req)
            except RequirementParseError as e:
                yield line_no, req, e


def requires_dists(location):
    # type: (Union[Distribution, MetadataFiles, Text]) -> Iterator[Requirement]
    """Examines dist for and returns any declared requirements.

    Looks for `Requires-Dist` metadata.

    The older `Requires` metadata is intentionally ignored, although we do log a warning if it is
    found to draw attention to this ~work-around and the associated issue in case any new data
    comes in.

    See:
    + https://www.python.org/dev/peps/pep-0345/#requires-dist-multiple-use
    + https://www.python.org/dev/peps/pep-0314/#requires-multiple-use

    :param location: A distribution to check for requirement metadata.
    :return: All requirements found.
    """
    if isinstance(location, Distribution):
        for requirement in location.metadata.requires_dists:
            yield requirement
        return

    metadata_files = None  # type: Optional[MetadataFiles]
    if isinstance(location, MetadataFiles):
        metadata_files = location
    else:
        # N.B.: This load can fail, but the source is the location of the metadata files, which
        # contains useful information in the path name for identifying the problem project and
        # version.
        metadata_files = load_metadata(location)
    if metadata_files is None:
        return

    invalid_values = []  # type: List[str]
    requires_dists = metadata_files.metadata.pkg_info.get_all("Requires-Dist", ())
    if not requires_dists and MetadataType.EGG_INFO is metadata_files.metadata.type:
        for metadata_file in "requires.txt", "depends.txt":
            content = metadata_files.read(metadata_file)
            if content:
                for requirement_or_error in _parse_requires_txt(content):
                    if isinstance(requirement_or_error, Requirement):
                        yield requirement_or_error
                    else:
                        line_no, req, err = requirement_or_error
                        invalid_values.append(
                            "{file}:{line_no} {req!r}: {err}".format(
                                file=(
                                    metadata_files.metadata_file_rel_path(metadata_file)
                                    or metadata_file
                                ),
                                line_no=line_no,
                                req=req,
                                err=err,
                            )
                        )
    else:
        for requires_dist in requires_dists:
            try:
                yield Requirement.parse(requires_dist)
            except RequirementParseError as e:
                invalid_values.append("{req!r}: {err}".format(req=requires_dist, err=e))
    if invalid_values:
        raise InvalidMetadataError(
            "Found {count} invalid Requires-Dist metadata {values} in {source}:\n"
            "{invalid_values}".format(
                count=len(invalid_values),
                values=pluralize(invalid_values, "value"),
                source=metadata_files.metadata.render_description(),
                invalid_values="\n".join(
                    "{index}. {invalid_value}".format(index=index, invalid_value=invalid_value)
                    for index, invalid_value in enumerate(invalid_values, start=1)
                ),
            )
        )

    legacy_requires = metadata_files.metadata.pkg_info.get_all("Requires", [])  # type: List[str]
    if legacy_requires:
        pex_warnings.warn(
            dedent(
                """\
                Ignoring {count} `Requires` {field} in {dist} metadata:
                {requires}

                You may have issues using the '{project_name}' distribution as a result.
                More information on this workaround can be found here:
                  https://github.com/pex-tool/pex/issues/1201#issuecomment-791715585
                """
            ).format(
                dist=location,
                project_name=metadata_files.metadata.project_name,
                count=len(legacy_requires),
                field=pluralize(legacy_requires, "field"),
                requires=os.linesep.join(
                    "{index}.) Requires: {req}".format(index=index, req=req)
                    for index, req in enumerate(legacy_requires, start=1)
                ),
            )
        )


class RequirementParseError(Exception):
    """Indicates and invalid requirement string.

    See PEP-508: https://www.python.org/dev/peps/pep-0508
    """


@attr.s(frozen=True)
class Constraint(object):
    @classmethod
    def parse(cls, constraint):
        # type: (Text) -> Constraint
        try:
            return cls.from_packaging_requirement(PackagingRequirement(constraint))
        except InvalidRequirement as e:
            raise RequirementParseError(str(e))

    @classmethod
    def from_packaging_requirement(cls, requirement):
        # type: (PackagingRequirement) -> Constraint
        return cls(
            name=requirement.name, specifier=requirement.specifier, marker=requirement.marker
        )

    name = attr.ib(eq=False)  # type: str
    specifier = attr.ib(factory=SpecifierSet)  # type: SpecifierSet
    marker = attr.ib(default=None, eq=str)  # type: Optional[Marker]

    project_name = attr.ib(init=False, repr=False)  # type: ProjectName
    _str = attr.ib(init=False, eq=False, repr=False)  # type: str
    _legacy_version = attr.ib(init=False, repr=False)  # type: Optional[str]

    def __attrs_post_init__(self):
        # type: () -> None
        object.__setattr__(self, "project_name", ProjectName(self.name))

        parts = [self.name]
        if self.specifier:
            parts.append(str(self.specifier))
        if self.marker:
            parts.append("; {marker}".format(marker=self.marker))
        object.__setattr__(self, "_str", "".join(parts))

        # We handle arbitrary equality separately since its semantics are simple - exact matches
        # only - and newer versions of packaging will fail to parse non PEP-440 compliant version
        # strings prior to performing the comparison. This needlessly negates the ~only useful
        # case for arbitrary equality - requiring exact legacy versions.
        specifiers = list(self.specifier)

        object.__setattr__(
            self,
            "_legacy_version",
            specifiers[0].version
            if len(specifiers) == 1 and "===" == specifiers[0].operator
            else None,
        )

    @property
    def key(self):
        # type: () -> str
        return self.project_name.normalized

    def __contains__(self, item):
        # type: (Union[str, Version, Distribution, ProjectNameAndVersion, Constraint]) -> bool

        # We emulate pkg_resources.Requirement.__contains__ pre-release behavior here since the
        # codebase expects it.
        return self.contains(item, prereleases=True)

    def contains(
        self,
        item,  # type: Union[str, Version, Distribution, ProjectNameAndVersion, Constraint]
        prereleases=None,  # type: Optional[bool]
    ):
        # type: (...) -> bool
        if isinstance(item, Constraint):
            return item.project_name == self.project_name and specifier_sets.includes(
                self.specifier, item.specifier
            )
        elif isinstance(item, ProjectNameAndVersion):
            if item.canonicalized_project_name != self.project_name:
                return False
            version = (
                item.canonicalized_version.raw
                if self._legacy_version
                else item.canonicalized_version.parsed_version
            )  # type: Union[ParsedVersion, Text]
        elif isinstance(item, Distribution):
            if item.key != self.key:
                return False
            version = (
                item.metadata.version.raw
                if self._legacy_version
                else item.metadata.version.parsed_version
            )
        elif isinstance(item, Version):
            version = item.raw if self._legacy_version else item.parsed_version
        else:
            version = item

        # N.B.: We handle the case of `===<legacy version>` specially since it easy to do
        # (arbitrary equality indicates an exact match) and packaging>=22.0 does not parse legacy
        # versions (even though it does handle `===`). Since the ~only useful case for `===` is
        # comparing legacy versions, this keeps that ability viable in a backwards compatible way
        # while still upgrading packaging past 22.0.
        if self._legacy_version:
            return version == self._legacy_version

        # We know SpecifierSet.contains returns bool on inspection of its code. The fact we import
        # via the pex.third_party mechanism makes the type opaque to MyPy. We also know it accepts
        # either a "parsed_version" or a str and take advantage of this to save re-parsing version
        # strings we've already parsed.
        return cast(bool, self.specifier.contains(version, prereleases=prereleases))

    def __str__(self):
        # type: () -> str
        return self._str

    def as_requirement(self):
        # type: () -> Requirement
        return Requirement(name=self.name, specifier=self.specifier, marker=self.marker)


@attr.s(frozen=True)
class Requirement(Constraint):
    @classmethod
    def parse(cls, requirement):
        # type: (Text) -> Requirement
        try:
            return cls.from_packaging_requirement(PackagingRequirement(requirement))
        except InvalidRequirement as e:
            raise RequirementParseError(str(e))

    @classmethod
    def from_packaging_requirement(cls, requirement):
        # type: (PackagingRequirement) -> Requirement
        return cls(
            name=requirement.name,
            url=requirement.url,
            extras=frozenset(requirement.extras),
            specifier=requirement.specifier,
            marker=requirement.marker,
        )

    url = attr.ib(default=None)  # type: Optional[str]
    extras = attr.ib(default=frozenset())  # type: FrozenSet[str]

    def __attrs_post_init__(self):
        # type: () -> None
        super(Requirement, self).__attrs_post_init__()

        parts = [self.name]
        if self.extras:
            parts.append("[{extras}]".format(extras=",".join(sorted(self.extras))))
        if self.specifier:
            parts.append(str(self.specifier))
        if self.url:
            parts.append("@ {url}".format(url=self.url))
            if self.marker:
                parts.append(" ")
        if self.marker:
            parts.append("; {marker}".format(marker=self.marker))
        object.__setattr__(self, "_str", "".join(parts))

    def as_constraint(self):
        # type: () -> Constraint
        return Constraint(name=self.name, specifier=self.specifier, marker=self.marker)


# N.B.: DistributionMetadata can have an expensive hash when a distribution has many requirements;
# so we cache the hash. See: https://github.com/pex-tool/pex/issues/1928
@attr.s(frozen=True, cache_hash=True)
class DistMetadata(object):
    @classmethod
    def from_metadata_files(cls, metadata_files):
        # type: (MetadataFiles) -> DistMetadata
        return cls(
            files=metadata_files,
            project_name=metadata_files.metadata.project_name,
            version=metadata_files.metadata.version,
            requires_dists=tuple(requires_dists(metadata_files)),
            requires_python=requires_python(metadata_files),
        )

    @classmethod
    def load(
        cls,
        location,  # type: Text
        *restrict_types_to  # type: MetadataType.Value
    ):
        # type: (...) -> DistMetadata

        metadata_files = load_metadata(location, restrict_types_to=restrict_types_to)
        if metadata_files is None:
            raise MetadataError(
                "Failed to determine project name and version for distribution at "
                "{location}.".format(location=location)
            )
        return cls.from_metadata_files(metadata_files)

    files = attr.ib(eq=False)  # type: MetadataFiles
    project_name = attr.ib()  # type: ProjectName
    version = attr.ib()  # type: Version
    requires_dists = attr.ib(default=())  # type: Tuple[Requirement, ...]
    requires_python = attr.ib(default=SpecifierSet())  # type: Optional[SpecifierSet]

    @property
    def type(self):
        # type: () -> MetadataType.Value
        return self.files.metadata.type


def _realpath(path):
    # type: (str) -> str
    return os.path.realpath(path)


class DistributionType(Enum["DistributionType.Value"]):
    class Value(Enum.Value):
        pass

    WHEEL = Value("whl")
    SDIST = Value("sdist")
    INSTALLED = Value("installed")

    @classmethod
    def of(cls, location):
        # type: (Text) -> DistributionType.Value
        if os.path.isdir(location):
            return cls.INSTALLED
        if location.endswith(".whl") and zipfile.is_zipfile(location):
            return cls.WHEEL
        return cls.SDIST


@attr.s(frozen=True)
class Distribution(object):
    @staticmethod
    def _read_metadata_lines(metadata_bytes):
        # type: (bytes) -> Iterator[Text]
        for line in metadata_bytes.splitlines():
            # This is pkg_resources.IMetadataProvider.get_metadata_lines behavior, which our
            # code expects.
            normalized = line.decode("utf-8").strip()
            if normalized and not normalized.startswith("#"):
                yield normalized

    @classmethod
    def parse_entry_map(cls, entry_points_contents):
        # type: (bytes) -> Dict[Text, Dict[Text, EntryPoint]]

        # This file format is defined here:
        #   https://packaging.python.org/en/latest/specifications/entry-points/#file-format

        entry_map = defaultdict(dict)  # type: DefaultDict[Text, Dict[Text, EntryPoint]]
        group = None  # type: Optional[Text]
        for index, line in enumerate(cls._read_metadata_lines(entry_points_contents), start=1):
            if line.startswith("[") and line.endswith("]"):
                group = line[1:-1]
            elif not group:
                raise ValueError(
                    "Failed to parse entry_points.txt, encountered an entry point with no "
                    "group on line {index}: {line}".format(index=index, line=line)
                )
            else:
                entry_point = EntryPoint.parse(line)
                entry_map[group][entry_point.name] = entry_point
        return entry_map

    @classmethod
    def load(cls, location):
        # type: (str) -> Distribution
        return cls(location=location, metadata=DistMetadata.load(location))

    # N.B.: Resolving the distribution location through any symlinks is pkg_resources behavior,
    # which our code expects.
    location = attr.ib(converter=_realpath)  # type: str

    metadata = attr.ib()  # type: DistMetadata

    @property
    def type(self):
        # type: () -> DistributionType.Value
        return DistributionType.of(self.location)

    @property
    def key(self):
        # type: () -> str
        return self.metadata.project_name.normalized

    @property
    def project_name(self):
        # type: () -> str
        return self.metadata.project_name.raw

    @property
    def version(self):
        # type: () -> str
        return self.metadata.version.raw

    def as_requirement(self):
        # type: () -> Requirement
        return Requirement(
            name=self.project_name,
            specifier=SpecifierSet(
                "{operator}{version}".format(
                    operator="===" if self.metadata.version.is_legacy else "==",
                    version=self.version,
                )
            ),
        )

    def requires(self):
        # type: () -> Tuple[Requirement, ...]
        return self.metadata.requires_dists

    def _read_metadata_file(self, name):
        # type: (str) -> Optional[bytes]
        normalized_name = os.path.normpath(name)
        if os.path.isabs(normalized_name):
            raise ValueError(
                "The metadata file name must be a relative path under the .dist-info/ (or "
                ".egg-info/) directory. Given: {name}".format(name=name)
            )

        return self.metadata.files.read(name)

    def iter_metadata_lines(self, name):
        # type: (str) -> Iterator[Text]
        contents = self._read_metadata_file(name)
        if contents:
            for line in self._read_metadata_lines(contents):
                yield line

    def get_entry_map(self):
        # type: () -> Dict[Text, Dict[Text, EntryPoint]]
        entry_points_metadata_file = self._read_metadata_file("entry_points.txt")
        if entry_points_metadata_file is None:
            return defaultdict(dict)
        return self.parse_entry_map(entry_points_metadata_file)

    def __str__(self):
        # type: () -> str
        return "{project_name} {version}".format(
            project_name=self.project_name, version=self.version
        )


@attr.s(frozen=True)
class EntryPoint(object):
    @classmethod
    def parse(cls, spec):
        # type: (Text) -> EntryPoint

        # This file format is defined here:
        #   https://packaging.python.org/en/latest/specifications/entry-points/#file-format

        components = spec.split("=")
        if len(components) != 2:
            raise ValueError("Invalid entry point specification: {spec}.".format(spec=spec))

        name, value = components
        # N.B.: Python identifiers must be ascii.
        module, sep, attrs = str(value).strip().partition(":")
        if sep and not attrs:
            raise ValueError("Invalid entry point specification: {spec}.".format(spec=spec))

        entry_point_name = name.strip()
        if sep:
            return CallableEntryPoint(
                name=entry_point_name, module=module, attrs=tuple(attrs.split("."))
            )

        return cls(name=entry_point_name, module=module)

    name = attr.ib()  # type: Text
    module = attr.ib()  # type: str

    def __str__(self):
        # type: () -> str
        return self.module


@attr.s(frozen=True)
class CallableEntryPoint(EntryPoint):
    _attrs = attr.ib()  # type: Tuple[str, ...]

    @_attrs.validator
    def _validate_attrs(self, _, value):
        if not value:
            raise ValueError("A callable entry point must select a callable item from the module.")

    def resolve(self):
        # type: () -> Callable[[], Any]
        module = importlib.import_module(self.module)
        try:
            return cast("Callable[[], Any]", functools.reduce(getattr, self._attrs, module))
        except AttributeError as e:
            raise ImportError(
                "Could not resolve {attrs} in {module}: {err}".format(
                    attrs=".".join(self._attrs), module=module, err=e
                )
            )

    def __str__(self):
        # type: () -> str
        return "{module}:{attrs}".format(module=self.module, attrs=".".join(self._attrs))


def find_distribution(
    project_name,  # type: Union[str, ProjectName]
    search_path=None,  # type: Optional[Iterable[str]]
    rescan=False,  # type: bool
):
    # type: (...) -> Optional[Distribution]

    canonicalized_project_name = (
        project_name if isinstance(project_name, ProjectName) else ProjectName(project_name)
    )
    for location in search_path or sys.path:
        if not os.path.isdir(location):
            continue
        metadata_files = load_metadata(
            location,
            project_name=canonicalized_project_name,
            restrict_types_to=(MetadataType.DIST_INFO, MetadataType.EGG_INFO),
            rescan=rescan,
        )
        if metadata_files:
            return Distribution(
                location=location, metadata=DistMetadata.from_metadata_files(metadata_files)
            )
    return None


def find_distributions(
    search_path=None,  # type: Optional[Iterable[str]]
    rescan=False,  # type: bool
):
    # type: (...) -> Iterator[Distribution]
    seen = set()
    for location in search_path or sys.path:
        if not os.path.isdir(location):
            continue
        for metadata_files in iter_metadata_files(
            location,
            restrict_types_to=(MetadataType.DIST_INFO, MetadataType.EGG_INFO),
            rescan=rescan,
        ):
            if metadata_files.metadata in seen:
                continue
            seen.add(metadata_files.metadata)
            yield Distribution(
                location=location, metadata=DistMetadata.from_metadata_files(metadata_files)
            )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import functools

from pex.pep_440 import Version
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterator, List, Optional, Tuple, Union

    import attr  # vendor:skip

else:
    from pex.third_party import attr


@attr.s(frozen=True)
class ArbitraryEquality(object):
    version = attr.ib()  # type: str


@functools.total_ordering
@attr.s(frozen=True, order=False)
class _Bound(object):
    version = attr.ib()  # type: Version
    inclusive = attr.ib()  # type: bool
    symbol = attr.ib(init=False, eq=False, repr=False)  # type: str
    rank = attr.ib(init=False, eq=False, repr=False)  # type: int

    def __attrs_post_init__(self):
        raise NotImplementedError(
            "_Bound must be subclassed and __attrs_post_init__ implemented to initialize symbol"
            "and rank."
        )

    def __lt__(self, other):
        # type: (Any) -> bool
        if type(self) != type(other):
            return NotImplemented
        return (self.version, self.rank) < (other.version, other.rank)

    def __str__(self):
        # type: () -> str
        parts = [self.symbol]
        if self.inclusive:
            parts.append("=")
        parts.append(str(self.version))
        return "".join(parts)


class LowerBound(_Bound):
    def __attrs_post_init__(self):
        # type: () -> None
        object.__setattr__(self, "symbol", ">")
        # I.E sorted order of: >=1, >1, >=2, >2
        object.__setattr__(self, "rank", 0 if self.inclusive else 1)


class UpperBound(_Bound):
    def __attrs_post_init__(self):
        # type: () -> None
        object.__setattr__(self, "symbol", "<")
        # I.E sorted order of: <1, <=1, <2, <=2
        object.__setattr__(self, "rank", 1 if self.inclusive else 0)


@attr.s(frozen=True)
class ExcludedRange(object):
    lower = attr.ib()  # type: LowerBound
    upper = attr.ib()  # type: UpperBound

    def intersects(self, other):
        # type: (Range) -> bool

        # Test partial overlap from below.
        if other.upper:
            if self.lower.version > other.upper.version:
                return False
            elif (
                self.lower.version == other.upper.version
                and self.lower.inclusive
                and other.upper.inclusive
            ):
                return True

        # Test partial overlap from above.
        if other.lower:
            if self.upper.version < other.lower.version:
                return False
            elif (
                self.upper.version == other.lower.version
                and self.upper.inclusive
                and other.lower.inclusive
            ):
                return True

        # Test full containment.
        excluded_range = Range(lower=self.lower, upper=self.upper)
        return other in excluded_range or excluded_range in other

    def __str__(self):
        return ",".join(map(str, (self.lower, self.upper)))


@attr.s(frozen=True)
class Range(object):
    lower = attr.ib(default=None)  # type: Optional[LowerBound]
    upper = attr.ib(default=None)  # type: Optional[UpperBound]
    excludes = attr.ib(default=())  # type: Tuple[ExcludedRange, ...]

    def __contains__(self, other):
        # type: (Any) -> bool
        if Range is not type(other):
            return NotImplemented
        return self.contains(other)

    def contains(self, other):
        # type: (Range) -> bool

        if self.lower:
            if not other.lower:
                return False
            elif self.lower.version > other.lower.version:
                return False
            elif (
                self.lower.version == other.lower.version
                and not self.lower.inclusive
                and other.lower.inclusive
            ):
                return False

        if self.upper:
            if not other.upper:
                return False
            elif self.upper.version < other.upper.version:
                return False
            elif (
                self.upper.version == other.upper.version
                and not self.upper.inclusive
                and other.upper.inclusive
            ):
                return False

        for excluded_range in self.excludes:
            if excluded_range.intersects(other):
                return False

        return True

    def __str__(self):
        # type: () -> str
        parts = []
        if self.lower:
            parts.append(str(self.lower))
        for excluded_range in self.excludes:
            parts.append("!({})".format(excluded_range))
        if self.upper:
            parts.append(str(self.upper))
        return ",".join(parts)


def _increment_release(
    version,  # type: Version
    index,  # type: int
):
    # type: (...) -> Version

    parsed_version = version.parsed_version

    parts = []  # type: List[str]
    if parsed_version.epoch != 0:
        parts.append(str(parsed_version.epoch))
        parts.append("!")
    if parsed_version.release:
        ceiling_release = list(parsed_version.release[:index])
        ceiling_release.append(parsed_version.release[index] + 1)
        parts.append(".".join(str(x) for x in ceiling_release))

    return Version("".join(parts))


def _bounds(specifier_set):
    # type: (SpecifierSet) -> Iterator[Union[ArbitraryEquality, ExcludedRange, LowerBound, UpperBound]]

    for spec in specifier_set:
        if "===" == spec.operator:
            yield ArbitraryEquality(spec.version)
            continue

        if spec.version.endswith(".*"):
            version = Version(spec.version[:-2])
            if spec.operator in ("==", "!="):
                lower = LowerBound(version, inclusive=True)
                # The SpecifierSet argument type ensures we have at least X.* across all vendored
                # versions of packaging we ship; so the -1 index access is safe.
                ceiling = _increment_release(version, index=-1)
                upper = UpperBound(ceiling, inclusive=False)
                if "==" == spec.operator:
                    yield lower
                    yield upper
                else:
                    yield ExcludedRange(lower=lower, upper=upper)
                continue
            # else:
            #   It's an error to use `.*` as a version suffix for anything other than `==` and
            #   `!=` operators in modern `packaging` but older versions do support it for the `>`,
            #   `>=`, `<` and `<=`, so we fall through for those knowing that if we're using a
            #   newer vendored packaging the SpecifierSet input will have already failed to parse,
            #   and we'll never get here.
        else:
            version = Version(spec.version)

        if "==" == spec.operator:
            yield LowerBound(version, inclusive=True)
            yield UpperBound(version, inclusive=True)
        elif "!=" == spec.operator:
            yield ExcludedRange(
                lower=LowerBound(version, inclusive=True),
                upper=UpperBound(version, inclusive=True),
            )
        elif ">=" == spec.operator:
            yield LowerBound(version, inclusive=True)
        elif ">" == spec.operator:
            yield LowerBound(version, inclusive=False)
        elif "<=" == spec.operator:
            yield UpperBound(version, inclusive=True)
        elif "<" == spec.operator:
            yield UpperBound(version, inclusive=False)
        elif "~=" == spec.operator:
            yield LowerBound(version, inclusive=True)
            # The SpecifierSet argument type ensures we have at least X.Y across all vendored
            # versions of packaging we ship; so the -2 index access is safe.
            ceiling = _increment_release(version, -2)
            yield UpperBound(ceiling, inclusive=False)


def as_range(specifier_set):
    # type: (Union[str, SpecifierSet]) -> Union[ArbitraryEquality, Range]

    lower_bounds = []  # type: List[LowerBound]
    upper_bounds = []  # type: List[UpperBound]
    excludes = []  # type: List[ExcludedRange]
    for bound in _bounds(
        specifier_set if isinstance(specifier_set, SpecifierSet) else SpecifierSet(specifier_set)
    ):
        if isinstance(bound, ArbitraryEquality):
            return bound
        elif isinstance(bound, ExcludedRange):
            excludes.append(bound)
        elif isinstance(bound, LowerBound):
            lower_bounds.append(bound)
        else:
            upper_bounds.append(bound)

    # Narrow bounds to the smallest range.
    lower = sorted(lower_bounds)[-1] if lower_bounds else None
    upper = sorted(upper_bounds)[0] if upper_bounds else None

    # Discard excludes outside the bounds and further narrow bounds by merging any overlapping
    # excludes.
    if lower:
        for exclude in sorted(excludes, key=lambda ex: ex.lower):
            # MyPy fails to typecheck <= under Python 2.7 only, even though LowerBound (via its
            # _Bound base class) has @total_ordering applied.
            if exclude.lower <= lower:  # type: ignore[operator]
                new_lower = LowerBound(exclude.upper.version, inclusive=not exclude.upper.inclusive)
                if new_lower > lower:
                    lower = new_lower
                excludes.remove(exclude)
    if upper:
        for exclude in sorted(excludes, key=lambda ex: ex.upper, reverse=True):
            # MyPy fails to typecheck >= under Python 2.7 only, even though UpperBound (via its
            # _Bound base class) has @total_ordering applied.
            if exclude.upper >= upper:  # type: ignore[operator]
                new_upper = UpperBound(exclude.lower.version, inclusive=not exclude.lower.inclusive)
                if new_upper < upper:
                    upper = new_upper
                excludes.remove(exclude)

    return Range(
        lower=lower,
        upper=upper,
        excludes=tuple(sorted(excludes, key=lambda ex: (ex.lower, ex.upper))),
    )


def includes(
    specifier,  # type: Union[str, SpecifierSet]
    candidate,  # type: Union[str, SpecifierSet]
):
    # type: (...) -> bool

    included_range = as_range(specifier)
    candidate_range = as_range(candidate)
    if isinstance(included_range, ArbitraryEquality) or isinstance(
        candidate_range, ArbitraryEquality
    ):
        return included_range == candidate_range
    return candidate_range in included_range



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import re
import subprocess
from glob import glob
from textwrap import dedent

from pex.common import is_exe, safe_open
from pex.pep_427 import install_wheel_interpreter
from pex.pip.installation import get_pip
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv
from testing import WheelBuilder, make_env

if TYPE_CHECKING:
    from typing import Any


def test_install_wheel_interpreter(tmpdir):
    # type: (Any) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir, install_pip=InstallationChoice.YES)
    cowsay_script = venv.bin_path("cowsay")
    assert not os.path.exists(cowsay_script)

    download_dir = os.path.join(str(tmpdir), "downloads")
    get_pip().spawn_download_distributions(
        download_dir=download_dir, requirements=["cowsay==5.0"]
    ).wait()

    wheel_dir = os.path.join(str(tmpdir), "wheels")
    get_pip().spawn_build_wheels(
        distributions=glob(os.path.join(download_dir, "*.tar.gz")), wheel_dir=wheel_dir
    ).wait()
    wheels = glob(os.path.join(wheel_dir, "*.whl"))
    assert 1 == len(wheels)
    cowsay_wheel = wheels[0]

    install_wheel_interpreter(cowsay_wheel, interpreter=venv.interpreter)
    assert is_exe(cowsay_script)
    assert b"5.0\n" == subprocess.check_output(args=[cowsay_script, "--version"])

    pip = venv.bin_path("pip")
    subprocess.check_call(args=[pip, "uninstall", "--yes", "cowsay"])
    assert not os.path.exists(cowsay_script)


def test_install_scripts(tmpdir):
    # type: (Any) -> None

    # N.B.: This example was taken from https://github.com/pypa/pip/issues/10661.
    top = os.path.join(str(tmpdir), "top")
    with safe_open(os.path.join(top, "mypackage", "__init__.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                def do_some_preprocessing():
                    print('Done some preprocessing')
                    print('Now starting an interactive session')
                """
            )
        )
    with safe_open(os.path.join(top, "scripts", "interactive_script.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                #!python -i

                import mypackage

                mypackage.do_some_preprocessing()
                """
            )
        )
    with safe_open(os.path.join(top, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from setuptools import setup

                setup(
                    name='mypackage',
                    version='0.0.1',
                    packages=['mypackage'],
                    scripts=['scripts/interactive_script.py']
                )
                """
            )
        )

    wheels = os.path.join(str(tmpdir), "wheels")
    wheel = WheelBuilder(source_dir=top, wheel_dir=wheels).bdist()

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir)
    script = venv.bin_path("interactive_script.py")
    assert not os.path.exists(script)

    install_wheel_interpreter(wheel, interpreter=venv.interpreter)
    assert is_exe(script)

    process = subprocess.Popen(
        args=[script],
        env=make_env(TERM=os.environ.get("TERM", "xterm")),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
    )
    output, _ = process.communicate()
    assert re.match(
        br"^Done some preprocessing\nNow starting an interactive session\n>>>>?$", output.strip()
    ), output



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os

try:
    from . import requires_python  # type:ignore[attr-defined] # This file will be relocated.

    python_full_versions = requires_python.PYTHON_FULL_VERSIONS
    python_versions = requires_python.PYTHON_VERSIONS
    python_majors = sorted(set(version[0] for version in python_full_versions))
except ImportError:
    python_full_versions = []
    python_versions = []
    python_majors = []

os_names = []
platform_systems = []
sys_platforms = []
platform_tag_regexps = []

# N.B.: The following environment variables are used by the Pex runtime to control Pip and must be
# kept in-sync with `locker.py`.
target_systems_file = os.environ.pop("_PEX_TARGET_SYSTEMS_FILE", None)

if target_systems_file:
    import json

    with open(target_systems_file) as fp:
        target_systems = json.load(fp)
    os_names = target_systems["os_names"]
    platform_systems = target_systems["platform_systems"]
    sys_platforms = target_systems["sys_platforms"]
    platform_tag_regexps = target_systems["platform_tag_regexps"]


def patch_marker_evaluate():
    from pip._vendor.packaging import markers  # type: ignore[import]

    from pex.exceptions import production_assert
    from pex.typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Callable, Dict, Iterable, List, Tuple, Type

    original_eval_op = markers._eval_op

    # N.B.: The `packaging` distribution vendored with Pip>=24.1 has a patch that tests the
    # `python_full_version` marker environment value to see if it ends with "+", in which case it
    # appends "local" to it (see: https://github.com/pypa/packaging/pull/802). We support this test
    # by ensuring our value for `python_full_version`, either a `skip` sentinel or a list value,
    # both support the `endswith(value)` test and always return `False`.
    class NeverEndsWithMixin(object):
        def endswith(self, _):
            # type: (str) -> bool
            return False

    skip = NeverEndsWithMixin()

    class EndsWithList(NeverEndsWithMixin, list):
        pass

    def versions_to_string(
        versions,  # type: Iterable[Tuple[int, ...]]
        version_modifier=lambda version: version,  # type: Callable[[str], str]
        list_type=list,  # type: Type[List]
    ):
        # type: (...) -> List[str]
        return list_type(version_modifier(".".join(map(str, version))) for version in versions)

    original_extra = markers.default_environment().get("extra")
    python_versions_strings = versions_to_string(python_versions) or skip
    python_full_versions_strings = (
        versions_to_string(
            python_full_versions,
            # N.B.: We replicate the logic `packaging` vendored with Pip>=24.1 has for dealing with
            # non-tagged CPython builds, which is fine to apply across all versions of Pip and its
            # vendored packaging that we support.
            version_modifier=lambda version: (
                version + "local" if version.endswith("+") else version
            ),
            list_type=EndsWithList,
        )
        or skip
    )
    os_names_strings = os_names or skip
    platform_systems_strings = platform_systems or skip
    sys_platforms_strings = sys_platforms or skip

    class EvaluationEnvironment(dict):
        @classmethod
        def get_env(
            cls,
            environment,  # type: Dict[Any, Any]
            name,  # type: Any
        ):
            # type: (...) -> Any
            production_assert(
                isinstance(environment, cls),
                "Expected environment to come from the {function} function, "
                "which we patch to return {expected_type}, but was {actual_type}".format(
                    function=markers.default_environment,
                    expected_type=cls,
                    actual_type=type(environment),
                ),
            )
            return environment[name]

        def __missing__(self, name):
            # type: (Any) -> Any
            if name == "extra":
                if original_extra is None:
                    raise markers.UndefinedEnvironmentName(
                        "'extra' does not exist in evaluation environment."
                    )
                return original_extra
            if name == "python_version":
                return python_versions_strings
            if name == "python_full_version":
                return python_full_versions_strings
            if name == "os_name":
                return os_names_strings
            if name == "platform_system":
                return platform_systems_strings
            if name == "sys_platform":
                return sys_platforms_strings
            return skip

    def _eval_op(
        lhs,  # type: Any
        op,  # type: Any
        rhs,  # type: Any
    ):
        # type: (...) -> bool
        if lhs is skip or rhs is skip:
            return True
        return any(
            original_eval_op(left, op, right)
            for left in (lhs if isinstance(lhs, list) else [lhs])
            for right in (rhs if isinstance(rhs, list) else [rhs])
        )

    # Works with all Pip vendored packaging distributions.
    markers.default_environment = EvaluationEnvironment
    # Covers Pip<24.1 vendored packaging.
    markers._get_env = EvaluationEnvironment.get_env

    markers._eval_op = _eval_op


def patch_wheel_model():
    from pip._internal.models.wheel import Wheel  # type: ignore[import]

    Wheel.support_index_min = lambda *args, **kwargs: 0

    supported_checks = [lambda *args, **kwargs: True]
    if python_versions:
        import re

        def supported_version(self, *_args, **_kwargs):
            if not hasattr(self, "_versions"):
                versions = set()
                abis = list(self.abis)
                is_abi3 = ["abi3"] == abis
                is_abi_none = ["none"] == abis
                for pyversion in self.pyversions:
                    # For the format, see: https://peps.python.org/pep-0425/#python-tag
                    match = re.search(r"^(?P<impl>\D{2,})(?P<major>\d)(?P<minor>\d+)?", pyversion)
                    if not match:
                        continue

                    impl = match.group("impl")
                    if impl not in ("cp", "pp", "py", "cpython", "pypy"):
                        continue

                    major = int(match.group("major"))
                    minor = match.group("minor")
                    if is_abi_none or (is_abi3 and major == 3):
                        versions.add(major)
                    elif minor:
                        versions.add((major, int(minor)))
                    else:
                        versions.add(major)

                self._versions = versions

            return any(
                (version in python_majors) or (version in python_versions)
                for version in self._versions
            )

        supported_checks.append(supported_version)

    if platform_tag_regexps:
        import re

        def supported_platform_tag(self, *_args, **_kwargs):
            if any(plat == "any" for plat in self.plats):
                return True
            for platform_tag_regexp in platform_tag_regexps:
                if any(re.search(platform_tag_regexp, plat) for plat in self.plats):
                    return True
            return False

        supported_checks.append(supported_platform_tag)

    Wheel.supported = lambda *args, **kwargs: all(
        check(*args, **kwargs) for check in supported_checks
    )

    # N.B.: This patch is a noop for the 20.3.4-patched Pip but is required in newer Pip.
    # The method is used as a speedup hack by newer Pip in some cases instead of
    # Wheel.support_index_min.
    Wheel.find_most_preferred_tag = lambda *args, **kwargs: 0


def patch():
    # 1.) Universal dependency environment marker applicability.
    #
    # Allows all dependencies in metadata to be followed regardless
    # of whether they apply to this system. For example, if this is
    # Python 3.10 but a marker says a dependency is only for
    # 'python_version < "3.6"' we still want to lock that dependency
    # subgraph too.
    patch_marker_evaluate()

    # 2.) Universal wheel tag applicability.
    #
    # Allows all wheel URLs to be checked even when the wheel does not
    # match system tags.
    patch_wheel_model()



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import re
import subprocess
import sys
from textwrap import dedent

import colors

from pex.targets import LocalInterpreter
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING
from testing import PY310, ensure_python_interpreter, run_pex_command

if TYPE_CHECKING:
    from typing import Any


def current_interpreter_applies(specifier):
    # type: (str) -> bool

    return LocalInterpreter.create().requires_python_applies(
        SpecifierSet(specifier), source=__name__ + ".current_interpreter_applies"
    )


def test_nominal(tmpdir):
    # type: (Any) -> None

    curl_exe = os.path.join(str(tmpdir), "curl.py")
    with open(curl_exe, "w") as fp:
        fp.write(
            dedent(
                """\
                # /// script
                # dependencies = ["requests"]
                # requires-python = ">=3.8,<3.13"
                # ///

                import io
                import os
                import sys
                from typing import NoReturn

                import requests


                def curl(url: str) -> None:
                    for chunk in requests.get(url, stream=True).iter_content(
                        chunk_size=io.DEFAULT_BUFFER_SIZE
                    ):
                        sys.stdout.buffer.write(chunk)


                def main() -> NoReturn:
                    if len(sys.argv) != 2:
                        sys.exit(f"Usage: {os.environ.get('PEX', sys.argv[0])} <URL>")
                    try:
                        curl(sys.argv[1])
                        sys.exit(0)
                    except requests.RequestException as e:
                        sys.exit(str(e))


                if __name__ == "__main__":
                    main()
                """
            )
        )
    curl_pex = os.path.join(str(tmpdir), "curl.pex")
    args = ["--exe", curl_exe, "-o", curl_pex]
    python = sys.executable
    if not current_interpreter_applies(">=3.8,<3.13"):
        python = ensure_python_interpreter(PY310)
        args.append("--python")
        args.append(python)

    run_pex_command(args=args).assert_success()
    output = (
        subprocess.check_output(args=[python, curl_pex, "https://docs.pex-tool.org"])
        .decode("utf-8")
        .strip()
    )
    assert output.startswith("<!doctype html>"), output
    assert output.endswith("</html>"), output


def test_dependencies_additive(tmpdir):
    # type: (Any) -> None

    cowsay_exe = os.path.join(str(tmpdir), "cowsay.py")
    with open(cowsay_exe, "w") as fp:
        fp.write(
            dedent(
                """\
                # /// script
                # dependencies = ["cowsay==5.0"]
                # ///

                from cowsay import main
                try:
                    from colors import yellow
                except ImportError:
                    yellow = lambda text: text

                if __name__ == "__main__":
                    main.tux(yellow("Bird Beak!"))
                """
            )
        )
    run_pex_command(args=["--exe", cowsay_exe]).assert_success(
        expected_output_re=r".*^{expected}$.*".format(expected=re.escape("| Bird Beak! |")),
        re_flags=re.DOTALL | re.MULTILINE,
    )
    run_pex_command(args=["--exe", cowsay_exe, "ansicolors==1.1.8"]).assert_success(
        expected_output_re=r".*^{expected}$.*".format(
            expected=re.escape("| {colorized} |".format(colorized=colors.yellow("Bird Beak!")))
        ),
        re_flags=re.DOTALL | re.MULTILINE,
    )


def test_dependencies_conflicting(tmpdir):
    # type: (Any) -> None

    cowsay_exe = os.path.join(str(tmpdir), "cowsay.py")
    with open(cowsay_exe, "w") as fp:
        fp.write(
            dedent(
                """\
                # /// script
                # dependencies = ["cowsay==5.0"]
                # ///
                """
            )
        )
    run_pex_command(
        args=["--exe", cowsay_exe, "cowsay>=6.0", "--resolver-version", "pip-2020-resolver"]
    ).assert_failure(
        expected_error_re=r".*{expected}.*".format(
            expected=re.escape(
                "ERROR: Cannot install cowsay==5.0 and cowsay>=6.0 because these package versions "
                "have conflicting dependencies."
            )
        ),
        re_flags=re.DOTALL,
    )


def test_targets_additive(tmpdir):
    # type: (Any) -> None

    hello_world_exe = os.path.join(str(tmpdir), "hello-world.py")
    with open(hello_world_exe, "w") as fp:
        fp.write(
            dedent(
                """\
                # /// script
                # requires-python = ">=3.10,<3.13"
                # ///

                import os
                import sys


                if __name__ == "__main__":
                    match sys.argv:
                        case [_, single_arg]:
                            print(f"Hello {single_arg}!")
                        case _:
                            sys.exit(f"Usage {os.environ.get('PEX', sys.argv[0])} <greetee>")
                """
            )
        )

    python = (
        ensure_python_interpreter(PY310)
        if not current_interpreter_applies(">=3.10,<3.13")
        else None
    )

    # In nominal mode it should work.
    run_pex_command(args=["--exe", hello_world_exe, "--", "World"], python=python).assert_success(
        expected_output_re=r"^Hello World!$"
    )

    # As should a compatible target addition.
    run_pex_command(
        args=[
            "--exe",
            hello_world_exe,
            "--platform",
            "macosx-10.9-x86_64-cp-310-cp310",
            "--",
            "Mac, I<3 you",
        ],
        python=python,
    ).assert_success(expected_output_re=r"^Hello Mac, I<3 you!$")

    # But an incompatible target addition should be detected and fail.
    run_pex_command(
        args=[
            "--exe",
            hello_world_exe,
            "--platform",
            "macosx-10.9-x86_64-cp-39-cp39",
            "--",
            "World",
        ],
        python=python,
        quiet=True,
    ).assert_failure(
        expected_error_re=re.escape(
            "The script metadata from {script} specifies a requires-python of <3.13,>=3.10 but "
            "the following configured targets are incompatible with that constraint: abbreviated "
            "platform cp39-cp39-macosx_10_9_x86_64".format(script=hello_world_exe)
        )
    )


def test_no_script_metadata(tmpdir):
    # type: (Any) -> None

    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                # /// script
                # dependencies = ["ansicolors==1.1.8"]
                # ///

                try:
                    from colors import yellow
                except ImportError:
                    yellow = lambda text: text

                if __name__ == "__main__":
                    print(yellow("Mellow"))
                """
            )
        )
    run_pex_command(args=["--exe", exe]).assert_success(
        expected_output_re=re.escape(colors.yellow("Mellow"))
    )
    run_pex_command(args=["--exe", exe, "--no-pep723"]).assert_success(expected_output_re=r"Mellow")



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.typing import TYPE_CHECKING
from testing import all_pythons, make_env, run_pex_command

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, List, Text, Tuple


@pytest.mark.parametrize(
    ["execution_mode_args"],
    [
        pytest.param([], id="zipapp"),
        pytest.param(["--venv"], id="venv"),
    ],
)
def test_execute(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=["cowsay==4.0", "-c", "cowsay", "-o", cowsay, "--sh-boot"] + execution_mode_args
    ).assert_success()
    assert "4.0" == subprocess.check_output(args=[cowsay, "--version"]).decode("utf-8").strip()


@pytest.mark.parametrize(
    ["execution_mode_args"],
    [
        pytest.param([], id="zipapp"),
        pytest.param(["--venv"], id="venv"),
    ],
)
def test_issue_1881(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    os.mkdir(pex_root)
    # Test that the runtime_pex_root is respected even when it is unwritable at build time.
    os.chmod(pex_root, 0o555)
    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=[
            "cowsay==4.0",
            "-c",
            "cowsay",
            "-o",
            cowsay,
            "--python-shebang",
            sys.executable,
            "--sh-boot",
            "--runtime-pex-root",
            pex_root,
        ]
        + execution_mode_args
    ).assert_success()
    # simulate pex_root writable at runtime.
    os.chmod(pex_root, 0o777)

    def _execute_cowsay_pex():
        return (
            subprocess.check_output(
                args=[cowsay, "--version"], env=make_env(PEX_VERBOSE=1), stderr=subprocess.STDOUT
            )
            .decode("utf-8")
            .strip()
        )

    # When this string is logged from the sh_boot script it indicates that the slow
    # path of running the zipapp via python interpreter taken.
    slow_path_output = "Running zipapp pex to lay itself out under PEX_ROOT."
    assert slow_path_output in _execute_cowsay_pex()
    assert slow_path_output not in _execute_cowsay_pex()


def interpreters():
    # type: () -> Iterable[Tuple[Text, List[Text]]]

    def iter_interpreters():
        # type: () -> Iterator[Tuple[Text, List[Text]]]

        def entry(path):
            # type: (Text) -> Tuple[Text, List[Text]]
            return os.path.basename(path), [path]

        yield entry(sys.executable)

        for interpreter in all_pythons():
            yield entry(interpreter)

        locations = (
            subprocess.check_output(
                args=["/usr/bin/env", "bash", "-c", "command -v ash bash busybox dash ksh sh zsh"]
            )
            .decode("utf-8")
            .splitlines()
        )
        for location in locations:
            basename = os.path.basename(location)
            if "busybox" == basename:
                yield "ash (via busybox)", [location, "ash"]
            else:
                yield entry(location)

    return sorted({name: args for name, args in iter_interpreters()}.items())


@pytest.mark.parametrize(
    ["interpreter_cmd"],
    [pytest.param(args, id=name) for name, args in interpreters()],
)
def test_execute_via_interpreter(
    tmpdir,  # type: Any
    interpreter_cmd,  # type: List[str]
):
    # type: (...) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=["cowsay==4.0", "-c", "cowsay", "-o", cowsay, "--sh-boot"]
    ).assert_success()

    assert (
        "4.0"
        == subprocess.check_output(args=interpreter_cmd + [cowsay, "--version"])
        .decode("utf-8")
        .strip()
    )


def test_python_shebang_respected(tmpdir):
    # type: (Any) -> None

    cowsay = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=[
            "cowsay==4.0",
            "-c",
            "cowsay",
            "-o",
            cowsay,
            "--sh-boot",
            "--python-shebang",
            # This is a strange shebang ~no-one would use since it short-circuits the PEX execution
            # to always just print the Python interpreter version, but it serves the purposes of:
            # 1. Proving our python shebang is honored by the bash boot.
            # 2. The bash boot treatment can handle shebangs with arguments in them.
            "{python} -V".format(python=sys.executable),
        ]
    ).assert_success()

    # N.B.: Python 2.7 does not send version to stdout; so we redirect stdout to stderr to be able
    # to uniformly retrieve the Python version.
    output = subprocess.check_output(args=[cowsay], stderr=subprocess.STDOUT).decode("utf-8")
    version = "Python {version}".format(version=".".join(map(str, sys.version_info[:3])))
    assert output.startswith(version), output


EXECUTION_MODE_ARGS_PERMUTATIONS = [
    pytest.param([], id="ZIPAPP"),
    pytest.param(["--venv"], id="VENV"),
    pytest.param(["--sh-boot"], id="ZIPAPP (--sh-boot)"),
    pytest.param(["--venv", "--sh-boot"], id="VENV (--sh-boot)"),
]


@pytest.mark.parametrize("execution_mode_args", EXECUTION_MODE_ARGS_PERMUTATIONS)
def test_issue_1782(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex.sh"))
    run_pex_command(
        args=[pex_project_dir, "-c", "pex", "-o", pex] + execution_mode_args
    ).assert_success()

    help_line1 = subprocess.check_output(args=[pex, "-h"]).decode("utf-8").splitlines()[0]
    assert help_line1.startswith("usage: {pex} ".format(pex=os.path.basename(pex))), help_line1
    assert (
        pex
        == subprocess.check_output(
            args=[pex, "-c", "import os; print(os.environ['PEX'])"], env=make_env(PEX_INTERPRETER=1)
        )
        .decode("utf-8")
        .strip()
    )


@pytest.mark.parametrize("execution_mode_args", EXECUTION_MODE_ARGS_PERMUTATIONS)
def test_argv0(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.realpath(os.path.join(str(tmpdir), "pex.sh"))
    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "app.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import os
                import sys


                def main():
                    print(json.dumps({"PEX": os.environ["PEX"], "argv0": sys.argv[0]}))


                if __name__ == "__main__":
                    main()
                """
            )
        )

    run_pex_command(
        args=["-D", src, "-e", "app:main", "-o", pex] + execution_mode_args
    ).assert_success()
    assert {"PEX": pex, "argv0": pex} == json.loads(subprocess.check_output(args=[pex]))

    run_pex_command(args=["-D", src, "-m", "app", "-o", pex] + execution_mode_args).assert_success()
    data = json.loads(subprocess.check_output(args=[pex]))
    assert pex == data.pop("PEX")
    assert "app.py" == os.path.basename(data.pop("argv0")), (
        "When executing modules we expect runpy.run_module to `alter_sys` in order to support "
        "pickling and other use cases as outlined in https://github.com/pex-tool/pex/issues/1018."
    )
    assert {} == data



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import subprocess
import sys

from pex.compatibility import to_unicode
from pex.typing import TYPE_CHECKING, cast
from testing import IntegResults

if TYPE_CHECKING:
    from typing import Text  # noqa
    from typing import Any


def run_pex3(
    *args,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> IntegResults

    python = cast("Text", kwargs.pop("python", to_unicode(sys.executable)))
    process = subprocess.Popen(
        args=[python, "-mpex.cli"] + list(args),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs
    )
    stdout, stderr = process.communicate()
    return IntegResults(
        output=stdout.decode("utf-8"), error=stderr.decode("utf-8"), return_code=process.returncode
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import errno
import json
import logging
import os
import re
import signal
import subprocess
import sys
import time

from pex.common import safe_open
from pex.typing import TYPE_CHECKING
from pex.variables import ENV
from pex.version import __version__

if TYPE_CHECKING:
    from typing import Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


logger = logging.getLogger(__name__)

SERVER_NAME = "Pex v{version} docs HTTP server".format(version=__version__)

_SERVER_DIR = os.path.join(ENV.PEX_ROOT, "docs", "server", __version__)


@attr.s(frozen=True)
class ServerInfo(object):
    url = attr.ib()  # type: str
    pid = attr.ib()  # type: int

    def __str__(self):
        # type: () -> str
        return "{url} @ {pid}".format(url=self.url, pid=self.pid)


@attr.s(frozen=True)
class Pidfile(object):
    _PIDFILE = os.path.join(_SERVER_DIR, "pidfile")

    @classmethod
    def load(cls):
        # type: () -> Optional[Pidfile]
        try:
            with open(cls._PIDFILE) as fp:
                data = json.load(fp)
            return cls(ServerInfo(url=data["url"], pid=data["pid"]))
        except (OSError, IOError, ValueError, KeyError) as e:
            logger.debug(
                "Failed to load {server} pid file from {path}: {err}".format(
                    server=SERVER_NAME, path=cls._PIDFILE, err=e
                )
            )
            return None

    @staticmethod
    def _read_url(
        server_log,  # type: str
        timeout,  # type: float
    ):
        # type: (...) -> Optional[str]

        # The permutations of Python versions, simple http server module and the output it provides:
        #  2.7: Serving HTTP on 0.0.0.0 port 46399 ... -mSimpleHttpServer
        #  3.5: Serving HTTP on 0.0.0.0 port 45577 ... -mhttp.server
        # 3.6+: Serving HTTP on 0.0.0.0 port 33539 (http://0.0.0.0:33539/) ... -mhttp.server

        start = time.time()
        while time.time() - start < timeout:
            with open(server_log) as fp:
                for line in fp:
                    if line.endswith(os.linesep):
                        match = re.search(r"Serving HTTP on 0\.0\.0\.0 port (?P<port>\d+)", line)
                        if match:
                            port = match.group("port")
                            return "http://localhost:{port}".format(port=port)
        return None

    @classmethod
    def record(
        cls,
        server_log,  # type: str
        pid,  # type: int
        timeout=5.0,  # type: float
    ):
        # type: (...) -> Optional[Pidfile]
        url = cls._read_url(server_log, timeout)
        if not url:
            return None

        with safe_open(cls._PIDFILE, "w") as fp:
            json.dump(dict(url=url, pid=pid), fp, indent=2, sort_keys=True)
        return cls(ServerInfo(url=url, pid=pid))

    server_info = attr.ib()  # type: ServerInfo

    def alive(self):
        # type: () -> bool
        # TODO(John Sirois): Handle pid rollover
        try:
            os.kill(self.server_info.pid, 0)
            return True
        except OSError as e:
            if e.errno == errno.ESRCH:  # No such process.
                return False
            raise

    def kill(self):
        # type: () -> None
        os.kill(self.server_info.pid, signal.SIGTERM)


@attr.s(frozen=True)
class LaunchResult(object):
    server_info = attr.ib()  # type: ServerInfo
    already_running = attr.ib()  # type: bool


# Frozen exception types don't work under 3.11+ where the `__traceback__` attribute can be set
# after construction in some cases.
@attr.s
class LaunchError(Exception):
    """Indicates an error launching the docs server."""

    log = attr.ib()  # type: str
    additional_msg = attr.ib(default=None)  # type: Optional[str]

    def __str__(self):
        # type: () -> str
        lines = ["Error launching docs server."]
        if self.additional_msg:
            lines.append(self.additional_msg)
        lines.append("See the log at {log} for more details.".format(log=self.log))
        return os.linesep.join(lines)


def launch(
    document_root,  # type: str
    port,  # type: int
    timeout=5.0,  # type: float
):
    # type: (...) -> LaunchResult

    pidfile = Pidfile.load()
    if pidfile and pidfile.alive():
        return LaunchResult(server_info=pidfile.server_info, already_running=True)

    # Not proper daemonization, but good enough.
    log = os.path.join(_SERVER_DIR, "log.txt")
    http_server_module = "http.server" if sys.version_info[0] == 3 else "SimpleHttpServer"
    env = os.environ.copy()
    # N.B.: We set up line buffering for the process pipes as well as the underlying Python running
    # the http server to ensure we can observe the `Serving HTTP on ...` line we need to grab the
    # ephemeral port chosen.
    env.update(PYTHONUNBUFFERED="1")
    with safe_open(log, "w") as fp:
        process = subprocess.Popen(
            args=[sys.executable, "-m", http_server_module, str(port)],
            env=env,
            cwd=document_root,
            preexec_fn=os.setsid,
            bufsize=1,
            stdout=fp.fileno(),
            stderr=subprocess.STDOUT,
        )

    pidfile = Pidfile.record(server_log=log, pid=process.pid, timeout=timeout)
    if not pidfile:
        try:
            os.kill(process.pid, signal.SIGKILL)
        except OSError as e:
            if e.errno != errno.ESRCH:  # No such process.
                raise LaunchError(
                    log,
                    additional_msg=(
                        "Also failed to kill the partially launched server at pid {pid}: "
                        "{err}".format(pid=process.pid, err=e)
                    ),
                )
        raise LaunchError(log)
    return LaunchResult(server_info=pidfile.server_info, already_running=False)


def shutdown():
    # type: () -> Optional[ServerInfo]

    pidfile = Pidfile.load()
    if not pidfile or not pidfile.alive():
        return None

    logger.debug("Killing {server} {info}".format(server=SERVER_NAME, info=pidfile.server_info))
    pidfile.kill()
    return pidfile.server_info



#!/usr/bin/env python3

import atexit
import hashlib
import io
import os
import shutil
import subprocess
import sys
import tempfile
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from email.parser import Parser
from enum import Enum, unique
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path, PurePath
from typing import Dict, Iterator, Optional, Tuple, cast

DIST_DIR = Path("dist")


def build_pex_pex(
    output_file: PurePath, verbosity: int = 0, env: Optional[Dict[str, str]] = None
) -> PurePath:
    # NB: We do not include the subprocess extra (which would be spelled: `.[subprocess]`) since we
    # would then produce a pex that would not be consumable by all python interpreters otherwise
    # meeting `python_requires`; ie: we'd need to then come up with a deploy environment / deploy
    # tooling, that built subprocess32 for linux cp27m, cp27mu, pypy, ... etc. Even with all the
    # work expended to do this, we'd still miss some platform someone wanted to run the Pex PEX on.
    # As such, we just ship unadorned Pex which is pure-python and universal. Any user wanting the
    # extra is encouraged to build a Pex PEX for their particular platform themselves.
    pex_requirement = "."

    args = [
        sys.executable,
        "-m",
        "pex",
        *["-v" for _ in range(verbosity)],
        "--disable-cache",
        "--no-build",
        "--no-compile",
        "--no-use-system-time",
        "--python-shebang",
        "/usr/bin/env python",
        "--no-strip-pex-env",
        "--include-tools",
        "-o",
        str(output_file),
        "-c",
        "pex",
        pex_requirement,
    ]
    subprocess.run(args=args, env=env, check=True)
    return output_file


def describe_rev() -> str:
    if not os.path.isdir(".git") and os.path.isfile("PKG-INFO"):
        # We're being build from an unpacked sdist.
        with open("PKG-INFO") as fp:
            return Parser().parse(fp).get("Version", "Unknown Version")

    git_describe = subprocess.run(
        ["git", "describe"], check=True, stdout=subprocess.PIPE, encoding="utf-8"
    )
    return git_describe.stdout.strip()


def describe_file(path: Path) -> Tuple[str, int]:
    hasher = hashlib.sha256()
    size = 0
    with path.open("rb") as fp:
        for chunk in iter(lambda: fp.read(io.DEFAULT_BUFFER_SIZE), b""):
            hasher.update(chunk)
            size += len(chunk)

    return hasher.hexdigest(), size


@unique
class Format(Enum):
    SDIST = "sdist"
    WHEEL = "wheel"

    def __str__(self) -> str:
        return cast(str, self.value)

    def build_arg(self) -> str:
        return f"--{self.value}"


def build_pex_dists(
    dist_fmt: Format,
    *additional_dist_fmts: Format,
    verbose: bool = False,
    env: Optional[Dict[str, str]] = None
) -> Iterator[PurePath]:
    tmp_dir = DIST_DIR / ".tmp"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    atexit.register(shutil.rmtree, tmp_dir, ignore_errors=True)
    out_dir = tempfile.mkdtemp(dir=tmp_dir)

    output = None if verbose else subprocess.DEVNULL

    subprocess.run(
        args=[
            sys.executable,
            "-m",
            "build",
            "--outdir",
            out_dir,
            *[fmt.build_arg() for fmt in [dist_fmt, *additional_dist_fmts]],
        ],
        env=env,
        stdout=output,
        stderr=output,
        check=True,
    )

    for dist in os.listdir(out_dir):
        built = DIST_DIR / dist
        shutil.move(os.path.join(out_dir, dist), built)
        yield built


def main(
    *additional_dist_formats: Format,
    verbosity: int = 0,
    embed_docs: bool = False,
    clean_docs: bool = False,
    pex_output_file: Optional[Path] = DIST_DIR / "pex",
    serve: bool = False
) -> None:
    env = os.environ.copy()
    if embed_docs:
        env.update(__PEX_BUILD_INCLUDE_DOCS__="1")

    if pex_output_file:
        print(f"Building Pex PEX to `{pex_output_file}` ...")
        build_pex_pex(pex_output_file, verbosity, env=env)

        rev = describe_rev()
        sha256, size = describe_file(pex_output_file)
        print(f"Built Pex PEX @ {rev}:")
        print(f"sha256: {sha256}")
        print(f"  size: {size}")

    if additional_dist_formats:
        print(
            f"Building additional distribution formats to `{DIST_DIR}`: "
            f'{", ".join(f"{i + 1}.) {fmt}" for i, fmt in enumerate(additional_dist_formats))} ...'
        )
        built = list(
            build_pex_dists(
                additional_dist_formats[0],
                *additional_dist_formats[1:],
                verbose=verbosity > 0,
                env=env
            )
        )
        print("Built:")
        for dist_path in built:
            print(f"  {dist_path}")

    if clean_docs:
        shutil.rmtree(DIST_DIR / "docs", ignore_errors=True)

    if serve:
        server = HTTPServer(("", 0), SimpleHTTPRequestHandler)
        host, port = server.server_address

        print(f"Serving Pex distributions from `{DIST_DIR}` at http://{host}:{port} ...")

        os.chdir(DIST_DIR)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print(f"Server shut down in response to keyboard interrupt.")


if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-v", dest="verbosity", action="count", default=0, help="Increase output verbosity level."
    )
    parser.add_argument(
        "--embed-docs",
        default=False,
        action="store_true",
        help="Embed offline docs in the built binary distributions.",
    )
    parser.add_argument(
        "--clean-docs",
        default=False,
        action="store_true",
        help="Clean up loose generated docs after they have embedded in binary distributions.",
    )
    parser.add_argument(
        "--additional-format",
        dest="additional_formats",
        choices=list(Format),
        type=Format,
        action="append",
        help="Package Pex in additional formats.",
    )
    parser.add_argument(
        "--no-pex",
        default=False,
        action="store_true",
        help="Do not build the Pex PEX.",
    )
    parser.add_argument(
        "--pex-output-file",
        default=DIST_DIR / "pex",
        type=Path,
        help="Build the Pex PEX at this path.",
    )
    parser.add_argument(
        "--serve",
        default=False,
        action="store_true",
        help="After packaging Pex serve up the packages over HTTP.",
    )
    args = parser.parse_args()

    main(
        *(args.additional_formats or ()),
        verbosity=args.verbosity,
        embed_docs=args.embed_docs,
        clean_docs=args.clean_docs,
        pex_output_file=None if args.no_pex else args.pex_output_file,
        serve=args.serve
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import shutil
import subprocess
import tarfile
from textwrap import dedent

import pytest

from pex.common import is_exe, safe_open
from pex.compatibility import urlparse
from pex.fetcher import URLFetcher
from pex.pep_440 import Version
from pex.pip.version import PipVersion
from pex.typing import TYPE_CHECKING
from testing import IS_LINUX, run_pex_command

if TYPE_CHECKING:
    from typing import Any


# TODO(John Sirois): Include a test of >= Pip 24.2 when Pex adds support for it.
#  See: https://github.com/pex-tool/pex/issues/2471
@pytest.mark.skipif(
    PipVersion.DEFAULT > PipVersion.VENDORED and PipVersion.DEFAULT.version < Version("24.2"),
    reason=(
        "Although Pex's vendored Pip is patched to handle statically linked musl libc CPython, no "
        "version of Pip Pex supports handles these Pythons until Pip 24.2"
    ),
)
@pytest.mark.skipif(
    not IS_LINUX,
    reason="This test tests statically linked musl libc CPython which is only available for Linux.",
)
def test_statically_linked_musl_libc_cpython_support(tmpdir):
    # type: (Any) -> None

    pbs_distribution_url = (
        "https://github.com/indygreg/python-build-standalone/releases/download/20221220/"
        "cpython-3.10.9+20221220-x86_64_v3-unknown-linux-musl-install_only.tar.gz"
    )
    pbs_distribution = os.path.join(
        str(tmpdir),
        os.path.basename(urlparse.urlparse(pbs_distribution_url).path),
    )
    with URLFetcher().get_body_stream(pbs_distribution_url) as read_fp, open(
        pbs_distribution, "wb"
    ) as write_fp:
        shutil.copyfileobj(read_fp, write_fp)
    with tarfile.open(pbs_distribution) as tf:
        tf.extractall(str(tmpdir))
    statically_linked_musl_libc_cpython = os.path.join(str(tmpdir), "python", "bin", "python3")
    assert is_exe(statically_linked_musl_libc_cpython)

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=["fortune==1.1.1", "-c", "fortune", "-o", pex],
        python=statically_linked_musl_libc_cpython,
    ).assert_success()

    fortune_db = os.path.join(str(tmpdir), "fortunes")
    with safe_open(fortune_db, "w") as fp:
        fp.write(
            dedent(
                """\
                A day for firm decisions!!!!!  Or is it?
                %
                """
            )
        )
    output = subprocess.check_output(args=[pex, fortune_db])
    assert b"A day for firm decisions!!!!!  Or is it?\n" == output, output.decode("utf-8")



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import glob
import os
import re
import shutil
from textwrap import dedent

import pytest

from pex.atomic_directory import atomic_directory
from pex.common import safe_open
from pex.compatibility import urlparse
from pex.dist_metadata import DistMetadata
from pex.pep_503 import ProjectName
from pex.pip.installation import get_pip
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve import resolver_configuration
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv
from testing import PY_VER, WheelBuilder, make_env, run_pex_command
from testing.mitmproxy import Proxy

if TYPE_CHECKING:
    from typing import Any, Iterable, Mapping

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class KeyringBackend(object):
    username = attr.ib()  # type: str
    password = attr.ib()  # type: str
    wheel = attr.ib()  # type: str
    project_name = attr.ib(init=False)  # type: ProjectName

    def __attrs_post_init__(self):
        # type: () -> None
        object.__setattr__(self, "project_name", DistMetadata.load(self.wheel).project_name)

    @property
    def basic_auth(self):
        # type: () -> str
        return "{username}:{password}".format(username=self.username, password=self.password)

    @property
    def requirement(self):
        # type: () -> str
        return "{project_name} @ file://{wheel}".format(
            project_name=self.project_name, wheel=self.wheel
        )


@pytest.fixture(scope="session")
def keyring_backend(shared_integration_test_tmpdir):
    # type: (str) -> KeyringBackend

    username = "jake"
    password = "jones"

    project_dir = os.path.join(shared_integration_test_tmpdir, "keyring-backend")
    with atomic_directory(project_dir) as atomic_projectdir:
        if not atomic_projectdir.is_finalized():
            with safe_open(
                os.path.join(atomic_projectdir.work_dir, "pex_test_backend.py"), "w"
            ) as fp:
                fp.write(
                    dedent(
                        """\
                        from keyring.backend import KeyringBackend
                        from keyring import credentials, errors


                        USERNAME = {username!r}
                        PASSWORD = {password!r}


                        class PexTestBackend(KeyringBackend):
                            priority = 9

                            def get_password(self, service, username):
                                return PASSWORD

                            def get_credential(self, service, username=None):
                                return credentials.SimpleCredential(username or USERNAME, PASSWORD)

                            def set_password(service, username, password):
                                raise errors.PasswordSetError("Unsupported.")

                            def delete_password(service, username):
                                raise errors.PasswordDeleteError("Unsupported.")
                        """
                    ).format(username=username, password=password)
                )

            with safe_open(os.path.join(atomic_projectdir.work_dir, "setup.py"), "w") as fp:
                fp.write(
                    dedent(
                        """\
                        from setuptools import setup


                        setup(
                            name="pex_test_backend",
                            version="0.1.0",
                            entry_points={
                                "keyring.backends": [
                                    "pex_test_backend = pex_test_backend",
                                ],
                            },
                            install_requires=["keyring==24.1.1"],
                            py_modules=["pex_test_backend"],
                        )
                        """
                    )
                )

            WheelBuilder(
                source_dir=atomic_projectdir.work_dir, wheel_dir=atomic_projectdir.work_dir
            ).bdist()

    wheels = glob.glob(os.path.join(project_dir, "*.whl"))
    assert 1 == len(wheels)
    return KeyringBackend(username=username, password=password, wheel=wheels[0])


@attr.s(frozen=True)
class KeyringVenv(object):
    backend = attr.ib()  # type: KeyringBackend
    path_element = attr.ib()  # type: str


@pytest.fixture(scope="session")
def keyring_venv(
    shared_integration_test_tmpdir,  # type: str
    keyring_backend,  # type: KeyringBackend
):
    # type: (...) -> KeyringVenv

    keyring_venv_dir = os.path.join(shared_integration_test_tmpdir, "keyring")
    with atomic_directory(keyring_venv_dir) as atomic_venvdir:
        if not atomic_venvdir.is_finalized():
            Virtualenv.create_atomic(
                venv_dir=atomic_venvdir,
                install_pip=InstallationChoice.YES,
                other_installs=[keyring_backend.requirement],
            )
    return KeyringVenv(backend=keyring_backend, path_element=Virtualenv(keyring_venv_dir).bin_dir)


@pytest.fixture
def index_url_info():
    # type: () -> urlparse.ParseResult
    index = os.environ.get("PIP_INDEX_URL", resolver_configuration.PYPI)
    return urlparse.urlparse(index)


@pytest.fixture
def index_reverse_proxy_target(index_url_info):
    # type: (urlparse.ParseResult) -> str
    return str(index_url_info._replace(path="", params="", query="", fragment="").geturl())


@pytest.fixture
def devpi_clean_env():
    # type: () -> Mapping[str, Any]

    # These will be set when tests are run with --devpi, and we want to unset them to
    # ensure our Pex command line config above is what is used.
    return dict(
        _PEX_USE_PIP_CONFIG=None,
        PIP_INDEX_URL=None,
        PIP_TRUSTED_HOST=None,
    )


skip_if_required_keyring_version_not_supported = pytest.mark.skipif(
    PY_VER < (3, 7), reason="The keyring distribution used for this test requires Python `>=3.7`."
)

keyring_provider_pip_versions = pytest.mark.parametrize(
    "pip_version",
    [
        pytest.param(pip_version, id=str(pip_version))
        for pip_version in PipVersion.values()
        # The Pip `--keyring-provider` option is only available starting in Pip 23.1.
        if pip_version >= PipVersion.v23_1 and pip_version.requires_python_applies()
    ],
)


def download_pip_requirements(
    download_dir,  # type: str
    pip_version,  # type: PipVersionValue
    extra_requirements=(),  # type: Iterable[str]
):
    # type: (...) -> None
    requirements = list(pip_version.requirements)
    requirements.extend(extra_requirements)
    get_pip(resolver=ConfiguredResolver.version(pip_version)).spawn_download_distributions(
        download_dir=download_dir, requirements=requirements
    ).wait()


@skip_if_required_keyring_version_not_supported
@keyring_provider_pip_versions
def test_subprocess_provider(
    proxy,  # type: Proxy
    pip_version,  # type: PipVersionValue
    keyring_venv,  # type: KeyringVenv
    index_url_info,  # type: urlparse.ParseResult
    index_reverse_proxy_target,  # type: str
    devpi_clean_env,  # type: Mapping[str, Any]
    tmpdir,  # type: Any
):
    # type: (...) -> None

    # N.B.: Pip subprocess keyring support presents a catch-22 for Pex unless there is a
    # non-authenticated source to resolve a non-vendored --pip-version from in the 1st place (since
    # the Pip subprocess keyring backend is only supported in non-vendored versions of Pip); so we
    # use a find-links repo pre-populated with the Pip version we need; just as a user would have
    # to.
    find_links = os.path.join(str(tmpdir), "find-links")
    download_pip_requirements(download_dir=find_links, pip_version=pip_version)

    with proxy.reverse(
        targets=[index_reverse_proxy_target], proxy_auth=keyring_venv.backend.basic_auth
    ) as (port, _):
        pex_root = os.path.join(str(tmpdir), "pex-root")
        proxied_index = str(
            index_url_info._replace(
                scheme="http",
                netloc="{username}@localhost:{port}".format(
                    username=keyring_venv.backend.username, port=port
                ),
            ).geturl()
        )
        run_pex_command(
            args=[
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--no-pypi",
                "--index",
                proxied_index,
                "--find-links",
                find_links,
                "--pip-version",
                str(pip_version),
                "--use-pip-config",
                "cowsay==5.0",
                "-c",
                "cowsay",
                "--",
                "Subprocess Auth!",
            ],
            env=make_env(
                PIP_KEYRING_PROVIDER="subprocess",
                PATH=os.pathsep.join(
                    (keyring_venv.path_element, os.environ.get("PATH", os.defpath))
                ),
                **devpi_clean_env
            ),
        ).assert_success(expected_output_re=r"^.*\| Subprocess Auth! \|.*$", re_flags=re.DOTALL)


@skip_if_required_keyring_version_not_supported
@keyring_provider_pip_versions
def test_import_provider(
    proxy,  # type: Proxy
    pip_version,  # type: PipVersionValue
    keyring_venv,  # type: KeyringVenv
    index_url_info,  # type: urlparse.ParseResult
    index_reverse_proxy_target,  # type: str
    devpi_clean_env,  # type: Mapping[str, Any]
    tmpdir,  # type: Any
):
    # type: (...) -> None

    # N.B.: Pip keyring support presents a catch-22 unless there is a non-authenticated source to
    # resolve the keyring distributions from in the 1st place; so we use a find-links repo
    # pre-populated with the keyring dependencies we need to authenticate; just as a user would
    # have to.
    find_links = os.path.join(str(tmpdir), "find-links")
    download_pip_requirements(
        download_dir=find_links,
        pip_version=pip_version,
        extra_requirements=[keyring_venv.backend.wheel],
    )
    shutil.copy(keyring_venv.backend.wheel, find_links)

    with proxy.reverse(
        targets=[index_reverse_proxy_target], proxy_auth=keyring_venv.backend.basic_auth
    ) as (port, _):
        pex_root = os.path.join(str(tmpdir), "pex-root")
        proxied_index = str(
            index_url_info._replace(
                scheme="http",
                netloc="localhost:{port}".format(port=port),
            ).geturl()
        )
        run_pex_command(
            args=[
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--no-pypi",
                "--index",
                proxied_index,
                "--find-links",
                find_links,
                "--extra-pip-requirement",
                str(keyring_venv.backend.project_name),
                "--pip-version",
                str(pip_version),
                "--use-pip-config",
                "cowsay==5.0",
                "-c",
                "cowsay",
                "--",
                "Import Auth!",
            ],
            env=make_env(PIP_KEYRING_PROVIDER="import", **devpi_clean_env),
        ).assert_success(expected_output_re=r"^.*\| Import Auth! \|.*$", re_flags=re.DOTALL)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys

import pytest

from pex.compatibility import commonpath
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv
from testing import IS_MAC, PY310, ensure_python_venv, run_pex_command

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 6),
    reason="The doit 0.34.2 distribution requires at least Python 3.6.",
)
def test_exclude(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "--pip-version",
            "latest",
            "--platform",
            "macosx_10.12-x86_64-cp-310-cp310",
            "--platform",
            "linux-x86_64-cp-310-cp310",
            "doit==0.34.2",
            "--exclude",
            "MacFSEvents",
            "--exclude",
            "pyinotify",
            "-c",
            "doit",
            "-o",
            pex,
            "--preserve-pip-download-log",
        ]
    ).assert_success()

    env = os.environ.copy()
    python, pip = ensure_python_venv(PY310)
    subprocess.check_call(args=[pip, "install", "MacFSEvents" if IS_MAC else "pyinotify"])
    env["PEX_INHERIT_PATH"] = "fallback"

    assert (
        b"0.34.2"
        == subprocess.check_output(args=[python, pex, "--version"], env=env).splitlines()[0]
    )

    venv = Virtualenv.enclosing(python)
    assert venv is not None
    env["PEX_INTERPRETER"] = "1"
    module_path = os.path.realpath(
        subprocess.check_output(
            args=[
                python,
                pex,
                "-c",
                "import {excluded_module}; print({excluded_module}.__file__)".format(
                    excluded_module="fsevents" if IS_MAC else "pyinotify"
                ),
            ],
            env=env,
        )
        .decode("utf=8")
        .strip()
    )
    assert venv.site_packages_dir == commonpath((venv.site_packages_dir, module_path)), module_path



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os

from pex.common import safe_mkdtemp
from pex.dependency_configuration import DependencyConfiguration
from pex.pip.download_observer import DownloadObserver, Patch, PatchSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Mapping, Optional


class PatchContext(object):
    _PEX_DEP_CONFIG_FILE_ENV_VAR_NAME = "_PEX_DEP_CONFIG_FILE"

    @classmethod
    def load_dependency_configuration(cls):
        # type: () -> DependencyConfiguration

        dep_config_file = os.environ.pop(cls._PEX_DEP_CONFIG_FILE_ENV_VAR_NAME)
        with open(dep_config_file) as fp:
            data = json.load(fp)
        return DependencyConfiguration.create(
            excluded=data["excluded"], overridden=data["overridden"]
        )

    @classmethod
    def dump_dependency_configuration(cls, dependency_configuration):
        # type: (DependencyConfiguration) -> Mapping[str, str]

        dep_config_file = os.path.join(safe_mkdtemp(), "dep_config.json")
        with open(dep_config_file, "w") as fp:
            json.dump(
                {
                    "excluded": [str(exclude) for exclude in dependency_configuration.excluded],
                    "overridden": [
                        str(override) for override in dependency_configuration.all_overrides()
                    ],
                },
                fp,
            )
        return {cls._PEX_DEP_CONFIG_FILE_ENV_VAR_NAME: dep_config_file}


def patch(dependency_configuration):
    # type: (DependencyConfiguration) -> Optional[DownloadObserver]

    if not dependency_configuration:
        return None

    return DownloadObserver(
        analyzer=None,
        patch_set=PatchSet.create(
            Patch.from_code_resource(
                __name__,
                "requires.py",
                **PatchContext.dump_dependency_configuration(dependency_configuration)
            )
        ),
    )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import glob
import json
import os.path
import re
import shutil
import subprocess
from os.path import commonprefix
from textwrap import dedent
from typing import Iterator

import pytest

from pex.common import safe_open
from pex.compatibility import PY2, commonpath
from pex.dist_metadata import Requirement
from pex.executor import Executor
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.pex_info import PexInfo
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.lockfile import json_codec
from pex.resolve.resolver_configuration import ResolverVersion
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv
from testing import PY_VER, data, make_env, run_pex_command
from testing.cli import run_pex3
from testing.lock import extract_lock_option_args, index_lock_artifacts

if TYPE_CHECKING:
    from typing import Any

    import attr  # vendor:skip
else:
    from pex.third_party import attr


REQUESTS_LOCK = data.path("locks", "requests.lock.json")


def assert_certifi_is_excluded(pex):
    # type: (str) -> None

    assert ["certifi"] == list(PexInfo.from_pex(pex).excluded)
    assert ProjectName("certifi") not in frozenset(
        dist.metadata.project_name for dist in PEX(pex).resolve()
    )


@pytest.fixture(scope="module")
def requests_certifi_excluded_pex(tmpdir_factory):
    # type: (Any) -> str

    pex_root = str(tmpdir_factory.mktemp("pex_root"))
    pex = str(tmpdir_factory.mktemp("pex"))
    run_pex_command(
        args=[
            "--lock",
            REQUESTS_LOCK,
            "--exclude",
            "certifi",
            "-o",
            pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()
    assert_certifi_is_excluded(pex)
    return pex


REQUESTS_CMD = [
    "-c",
    "import os, requests, sys; print(os.path.realpath(sys.modules['certifi'].__file__))",
]
EXPECTED_IMPORT_ERROR_MSG = "ModuleNotFoundError: No module named 'certifi'"


@pytest.fixture(scope="module")
def certifi_venv(tmpdir_factory):
    # type: (Any) -> Virtualenv

    venv = Virtualenv.create(
        venv_dir=str(tmpdir_factory.mktemp("venv")), install_pip=InstallationChoice.YES
    )
    pip = venv.bin_path("pip")

    # N.B.: The constraining lock requirement is the one expressed by requests: certifi>=2017.4.17
    # The actual locked version is 2023.7.22; so we stress this crease and use a different, but
    # allowed, version.
    subprocess.check_call(args=[pip, "install", "certifi==2017.4.17"])

    return venv


skip_unless_compatible_with_requests_lock = pytest.mark.skipif(
    PY_VER < (3, 7) or PY_VER >= (3, 13), reason="The lock used is for >=3.7,<3.13"
)


def assert_certifi_import_behavior(
    pex,  # type: str
    certifi_venv,  # type: Virtualenv
):
    requests_cmd = [pex] + REQUESTS_CMD

    # Although the venv has certifi available, a PEX is hermetic by default; so it shouldn't be
    # used.
    with pytest.raises(Executor.NonZeroExit) as exc:
        certifi_venv.interpreter.execute(args=requests_cmd)
    assert EXPECTED_IMPORT_ERROR_MSG in exc.value.stderr

    # Allowing the `sys.path` to be inherited should allow the certifi hole to be filled in.
    _, stdout, _ = certifi_venv.interpreter.execute(
        args=requests_cmd, env=make_env(PEX_INHERIT_PATH="fallback")
    )
    assert certifi_venv.site_packages_dir == commonprefix(
        [certifi_venv.site_packages_dir, stdout.strip()]
    )


def assert_requests_certifi_excluded_pex(
    pex,  # type: str
    certifi_venv,  # type: Virtualenv
):
    # type: (...) -> None

    requests_cmd = [pex] + REQUESTS_CMD

    # The exclude option is buyer beware. A PEX using this option will not work if the excluded
    # distributions carry modules that are, in fact, needed at run time.
    process = subprocess.Popen(args=requests_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    _, stderr = process.communicate()
    assert process.returncode != 0
    assert EXPECTED_IMPORT_ERROR_MSG in stderr.decode("utf-8"), stderr.decode("utf-8")

    assert_certifi_import_behavior(pex, certifi_venv)


@skip_unless_compatible_with_requests_lock
def test_exclude(
    tmpdir,  # type: Any
    requests_certifi_excluded_pex,  # type: str
    certifi_venv,  # type: Virtualenv
):
    # type: (...) -> None

    assert_requests_certifi_excluded_pex(requests_certifi_excluded_pex, certifi_venv)


@skip_unless_compatible_with_requests_lock
def test_requirements_pex_exclude(
    tmpdir,  # type: Any
    requests_certifi_excluded_pex,  # type: str
    certifi_venv,  # type: Virtualenv
):
    # type: (...) -> None

    pex_root = PexInfo.from_pex(requests_certifi_excluded_pex).pex_root
    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "--requirements-pex",
            requests_certifi_excluded_pex,
            "ansicolors==1.1.8",
            "-o",
            pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()

    # Shouldn't need the certifi hole filled to import colors.
    output = subprocess.check_output(args=[pex, "-c", "import colors; print(colors.__file__)"])
    assert pex_root == commonprefix([pex_root, output.decode("utf-8").strip()])

    assert_certifi_import_behavior(pex, certifi_venv)


@skip_unless_compatible_with_requests_lock
def test_lock_exclude(
    tmpdir,  # type: Any
    certifi_venv,  # type: Virtualenv
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    shutil.copy(REQUESTS_LOCK, lock)

    run_pex3(
        *(["lock", "sync", "--exclude", "certifi", "--lock", lock] + extract_lock_option_args(lock))
    ).assert_success(
        expected_error_re=r"^.*{expected_message}.*$".format(
            expected_message=re.escape(
                dedent(
                    """\
                    Updates for lock generated by universal:
                      Deleted certifi 2023.7.22
                    """
                )
            ),
        ),
        re_flags=re.DOTALL,
    )

    lockfile = json_codec.load(lock)
    assert SortedTuple([Requirement.parse("certifi")]) == lockfile.excluded
    assert ProjectName("certifi") not in index_lock_artifacts(lockfile)

    pex = os.path.join(str(tmpdir), "pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")
    run_pex_command(
        args=[
            "--lock",
            lock,
            "-o",
            pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()

    assert_certifi_is_excluded(pex)
    assert_requests_certifi_excluded_pex(pex, certifi_venv)


@attr.s(frozen=True)
class PipOptions(object):
    @classmethod
    def iter(cls):
        # type: () -> Iterator[PipOptions]
        for pip_version in PipVersion.values():
            if not pip_version.requires_python_applies():
                continue
            for resolver_version in ResolverVersion.values():
                if not ResolverVersion.applies(resolver_version, pip_version):
                    continue
                yield cls(pip_version=pip_version, resolver_version=resolver_version)

    pip_version = attr.ib()  # type: PipVersionValue
    resolver_version = attr.ib()  # type: ResolverVersion.Value

    def iter_args(self):
        # type: () -> Iterator[str]
        yield "--pip-version"
        yield str(self.pip_version)
        yield "--resolver-version"
        yield str(self.resolver_version)

    def __str__(self):
        # type: () -> str
        return "-".join(
            (
                str(self.pip_version),
                "legacy" if self.resolver_version is ResolverVersion.PIP_LEGACY else "resolvelib",
            )
        )


@pytest.mark.parametrize(
    "pip_options",
    [pytest.param(pip_options, id=str(pip_options)) for pip_options in PipOptions.iter()],
)
def test_exclude_deep(
    tmpdir,  # type: Any
    pip_options,  # type: PipOptions
):
    # type: (...) -> None

    # Bootstrap the Pip version being used if needed before we turn off PyPI.
    if pip_options.pip_version is not PipVersion.VENDORED:
        run_pex_command(
            args=list(pip_options.iter_args()) + ["ansicolors==1.1.8", "--", "-c", ""]
        ).assert_success()

    venv = Virtualenv.create(
        os.path.join(str(tmpdir), "venv"),
        install_pip=InstallationChoice.UPGRADED,
        install_setuptools=InstallationChoice.UPGRADED,
        install_wheel=InstallationChoice.UPGRADED,
    )
    pip = venv.bin_path("pip")

    find_links = os.path.join(str(tmpdir), "find_links")
    project_dir = os.path.join(str(tmpdir), "projects")

    foo_dir = os.path.join(project_dir, "foo")
    with safe_open(os.path.join(foo_dir, "foo.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from bar import BAR


                def foo():
                    return BAR * 42
                """
            )
        )
    with safe_open(os.path.join(foo_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from setuptools import setup


                setup(
                    name="foo",
                    version="0.1.0",
                    install_requires=["bar"],
                    py_modules=["foo"],
                )
                """
            )
        )
    venv.interpreter.execute(
        args=["setup.py", "bdist_wheel", "--dist-dir", find_links], cwd=foo_dir
    )

    bar_dir = os.path.join(project_dir, "bar")
    with safe_open(os.path.join(bar_dir, "bar.py"), "w") as fp:
        print("BAR=1", file=fp)
    with safe_open(os.path.join(bar_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import os
                import sys

                from setuptools import setup


                if "BEHAVE" not in os.environ:
                    sys.exit("I'm an evil package.")


                setup(
                    name="bar",
                    version="0.1.0",
                    py_modules=["bar"],
                )
                """
            )
        )

    def assert_stderr_contains(
        expected,  # type: bytes
        *args,  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        process = subprocess.Popen(args=args, stderr=subprocess.PIPE, **kwargs)
        _, stderr = process.communicate()
        assert process.returncode != 0
        assert expected in stderr, stderr.decode()

    # The bar package should aggressively blow up in normal circumstances.
    assert_stderr_contains(
        b"I'm an evil package.",
        venv.interpreter.binary,
        "setup.py",
        "bdist_wheel",
        "--dist-dir",
        os.path.join(bar_dir, "dist"),
        cwd=bar_dir,
    )

    venv.interpreter.execute(
        args=["setup.py", "bdist_wheel", "--dist-dir", os.path.join(bar_dir, "dist")],
        cwd=bar_dir,
        env=make_env(BEHAVE=1),
    )
    wheels = glob.glob(os.path.join(bar_dir, "dist", "*.whl"))
    assert len(wheels) == 1
    bar_whl = wheels[0]

    venv.interpreter.execute(
        args=["setup.py", "sdist", "--dist-dir", find_links], cwd=bar_dir, env=make_env(BEHAVE=1)
    )

    # Building a Pex that requires bar should (transitively) aggressively blow up in normal
    # circumstances.
    run_pex_command(
        args=list(pip_options.iter_args()) + ["-f", find_links, "--no-pypi", "foo"]
    ).assert_failure(expected_error_re=r".*I'm an evil package\..*", re_flags=re.DOTALL)

    # But an `--exclude bar` should solve this by never resolving bar at all.
    exe = os.path.join(str(tmpdir), "exe.py")
    with safe_open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import os
                import sys

                import bar
                from foo import foo


                json.dump({"foo": foo(), "bar": os.path.realpath(bar.__file__)}, sys.stdout)
                """
            )
        )
    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=list(pip_options.iter_args())
        + ["-f", find_links, "--no-pypi", "foo", "--exclude", "bar", "-o", pex, "--exe", exe]
    ).assert_success()

    # The `--exclude bar` should hobble the PEX by default though, since bar is needed but missing.
    assert_stderr_contains(
        b"ImportError: No module named bar"
        if PY2
        else b"ModuleNotFoundError: No module named 'bar'",
        pex,
    )

    # But leaking in an externally installed bar should solve things.
    subprocess.check_call(args=[pip, "install", bar_whl])
    data = json.loads(
        subprocess.check_output(args=[pex], env=make_env(PEX_EXTRA_SYS_PATH=venv.site_packages_dir))
    )
    assert 42 == data.pop("foo")
    bar_module_path = data.pop("bar")
    assert venv.site_packages_dir == commonpath(
        (venv.site_packages_dir, bar_module_path)
    ), bar_module_path
    assert not data
